import json
from unittest.mock import patch

import pytest
import responses

from cumulusci.core.exceptions import CumulusCIException
from cumulusci.tasks.salesforce.tests.util import create_task
from cumulusci.tasks.salesforce.users.permsets import (
    AssignPermissionSetGroups,
    AssignPermissionSetLicenses,
    AssignPermissionSets,
)
from cumulusci.tests.util import CURRENT_SF_API_VERSION


class TestCreatePermissionSet:
    @responses.activate
    def test_create_permset(self):
        task = create_task(
            AssignPermissionSets,
            {
                "api_names": "PermSet1,PermSet2",
            },
        )

        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2C%28SELECT+PermissionSetId+FROM+PermissionSetAssignments%29+FROM+User+WHERE+Username+%3D+%27test-cci%40example.com%27",
            status=200,
            json={
                "done": True,
                "totalSize": 1,
                "records": [
                    {
                        "Id": "005000000000000",
                        "PermissionSetAssignments": {
                            "done": True,
                            "totalSize": 1,
                            "records": [{"PermissionSetId": "0PS000000000000"}],
                        },
                    }
                ],
            },
        )
        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2CName+FROM+PermissionSet+WHERE+Name+IN+%28%27PermSet1%27%2C+%27PermSet2%27%29",
            status=200,
            json={
                "done": True,
                "totalSize": 2,
                "records": [
                    {
                        "Id": "0PS000000000000",
                        "Name": "PermSet1",
                    },
                    {
                        "Id": "0PS000000000001",
                        "Name": "PermSet2",
                    },
                ],
            },
        )
        responses.add(
            method="POST",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/composite/sobjects",
            status=200,
            json=[{"id": "0Pa000000000001", "success": True, "errors": []}],
            match=[
                responses.json_params_matcher(
                    {
                        "allOrNone": False,
                        "records": [
                            {
                                "attributes": {"type": "PermissionSetAssignment"},
                                "AssigneeId": "005000000000000",
                                "PermissionSetId": "0PS000000000001",
                            }
                        ],
                    }
                ),
            ],
        )

        task()

        assert len(responses.calls) == 3

    @responses.activate
    def test_create_permset__alias(self):
        task = create_task(
            AssignPermissionSets,
            {
                "api_names": "PermSet1,PermSet2",
                "user_alias": "test0,test1",
            },
        )

        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2C%28SELECT+PermissionSetId+FROM+PermissionSetAssignments%29+FROM+User+WHERE+Alias+IN+%28%27test0%27%2C%27test1%27%29",
            status=200,
            json={
                "done": True,
                "totalSize": 2,
                "records": [
                    {
                        "Id": "005000000000000",
                        "PermissionSetAssignments": {
                            "done": True,
                            "totalSize": 1,
                            "records": [{"PermissionSetId": "0PS000000000000"}],
                        },
                    },
                    {
                        "Id": "005000000000001",
                        "PermissionSetAssignments": {
                            "done": True,
                            "totalSize": 1,
                            "records": [{"PermissionSetId": "0PS000000000000"}],
                        },
                    },
                ],
            },
        )
        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2CName+FROM+PermissionSet+WHERE+Name+IN+%28%27PermSet1%27%2C+%27PermSet2%27%29",
            status=200,
            json={
                "done": True,
                "totalSize": 1,
                "records": [
                    {
                        "Id": "0PS000000000000",
                        "Name": "PermSet1",
                    },
                    {
                        "Id": "0PS000000000001",
                        "Name": "PermSet2",
                    },
                ],
            },
        )
        responses.add(
            method="POST",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/composite/sobjects",
            status=200,
            json=[
                {"id": "0Pa000000000000", "success": True, "errors": []},
                {"id": "0Pa000000000001", "success": True, "errors": []},
            ],
            match=[
                responses.json_params_matcher(
                    {
                        "allOrNone": False,
                        "records": [
                            {
                                "attributes": {"type": "PermissionSetAssignment"},
                                "AssigneeId": "005000000000000",
                                "PermissionSetId": "0PS000000000001",
                            },
                            {
                                "attributes": {"type": "PermissionSetAssignment"},
                                "AssigneeId": "005000000000001",
                                "PermissionSetId": "0PS000000000001",
                            },
                        ],
                    }
                ),
            ],
        )

        task()

        assert len(responses.calls) == 3

    @responses.activate
    def test_create_permset__alias_raises(self):
        task = create_task(
            AssignPermissionSets,
            {
                "api_names": "PermSet1,PermSet2",
                "user_alias": "test",
            },
        )

        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2C%28SELECT+PermissionSetId+FROM+PermissionSetAssignments%29+FROM+User+WHERE+Alias+IN+%28%27test%27%29",
            status=200,
            json={
                "done": True,
                "totalSize": 0,
                "records": [],
            },
        )
        with pytest.raises(CumulusCIException):
            task()

    @responses.activate
    def test_create_permset__split_requests(self):
        task = create_task(
            AssignPermissionSets,
            {
                "api_names": ",".join(["PermSet" + str(i) for i in range(20)]),
                "user_alias": ",".join(["test" + str(i) for i in range(20)]),
            },
        )

        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2C%28SELECT+PermissionSetId+FROM+PermissionSetAssignments%29+FROM+User+WHERE+Alias+IN+%28%27test0%27%2C%27test1%27%2C%27test2%27%2C%27test3%27%2C%27test4%27%2C%27test5%27%2C%27test6%27%2C%27test7%27%2C%27test8%27%2C%27test9%27%2C%27test10%27%2C%27test11%27%2C%27test12%27%2C%27test13%27%2C%27test14%27%2C%27test15%27%2C%27test16%27%2C%27test17%27%2C%27test18%27%2C%27test19%27%29",
            status=200,
            json={
                "done": True,
                "totalSize": 20,
                "records": [
                    {
                        "Id": f"00500000000000{str(i)}",
                        "PermissionSetAssignments": None,
                    }
                    for i in range(20)
                ],
            },
        )
        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2CName+FROM+PermissionSet+WHERE+Name+IN+%28%27PermSet0%27%2C+%27PermSet1%27%2C+%27PermSet2%27%2C+%27PermSet3%27%2C+%27PermSet4%27%2C+%27PermSet5%27%2C+%27PermSet6%27%2C+%27PermSet7%27%2C+%27PermSet8%27%2C+%27PermSet9%27%2C+%27PermSet10%27%2C+%27PermSet11%27%2C+%27PermSet12%27%2C+%27PermSet13%27%2C+%27PermSet14%27%2C+%27PermSet15%27%2C+%27PermSet16%27%2C+%27PermSet17%27%2C+%27PermSet18%27%2C+%27PermSet19%27%29",
            status=200,
            json={
                "done": True,
                "totalSize": 20,
                "records": [
                    {
                        "Id": f"0PS000000000000{str(i)}",
                        "Name": f"PermSet{str(i)}",
                    }
                    for i in range(20)
                ],
            },
        )
        responses.add(
            method="POST",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/composite/sobjects",
            status=200,
            json=[
                {"id": f"0Pa00000000000{str(i)}", "success": True, "errors": []}
                for i in range(200)
            ],
        )

        task()

        assert len(responses.calls) == 4
        assert len(json.loads(responses.calls[2].request.body)["records"]) == 200
        assert len(json.loads(responses.calls[3].request.body)["records"]) == 200

    @responses.activate
    def test_create_permset_raises(self):
        task = create_task(
            AssignPermissionSets,
            {
                "api_names": "PermSet1,PermSet2,PermSet3",
            },
        )

        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2C%28SELECT+PermissionSetId+FROM+PermissionSetAssignments%29+FROM+User+WHERE+Username+%3D+%27test-cci%40example.com%27",
            status=200,
            json={
                "done": True,
                "totalSize": 1,
                "records": [
                    {
                        "Id": "005000000000000",
                        "PermissionSetAssignments": {
                            "done": True,
                            "totalSize": 1,
                            "records": [{"PermissionSetId": "0PS000000000000"}],
                        },
                    }
                ],
            },
        )
        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2CName+FROM+PermissionSet+WHERE+Name+IN+%28%27PermSet1%27%2C+%27PermSet2%27%2C+%27PermSet3%27%29",
            status=200,
            json={
                "done": True,
                "totalSize": 1,
                "records": [
                    {
                        "Id": "0PS000000000000",
                        "Name": "PermSet1",
                    },
                    {
                        "Id": "0PS000000000001",
                        "Name": "PermSet2",
                    },
                ],
            },
        )

        with pytest.raises(CumulusCIException):
            task()

    @responses.activate
    @patch("cumulusci.tasks.salesforce.users.permsets.CliTable", autospec=True)
    def test_create_permset_partial_success_raises(self, table):
        task = create_task(
            AssignPermissionSets,
            {
                "api_names": "PermSet1,PermSet2",
                "user_alias": "test0,test1",
            },
        )

        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2C%28SELECT+PermissionSetId+FROM+PermissionSetAssignments%29+FROM+User+WHERE+Alias+IN+%28%27test0%27%2C%27test1%27%29",
            status=200,
            json={
                "done": True,
                "totalSize": 2,
                "records": [
                    {
                        "Id": "005000000000000",
                        "PermissionSetAssignments": {
                            "done": True,
                            "totalSize": 1,
                            "records": [{"PermissionSetId": "0PS000000000000"}],
                        },
                    },
                    {
                        "Id": "005000000000001",
                        "PermissionSetAssignments": {
                            "done": True,
                            "totalSize": 1,
                            "records": [{"PermissionSetId": "0PS000000000000"}],
                        },
                    },
                ],
            },
        )
        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2CName+FROM+PermissionSet+WHERE+Name+IN+%28%27PermSet1%27%2C+%27PermSet2%27%29",
            status=200,
            json={
                "done": True,
                "totalSize": 1,
                "records": [
                    {
                        "Id": "0PS000000000000",
                        "Name": "PermSet1",
                    },
                    {
                        "Id": "0PS000000000001",
                        "Name": "PermSet2",
                    },
                ],
            },
        )
        responses.add(
            method="POST",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/composite/sobjects",
            status=200,
            json=[
                {"id": "0Pa000000000000", "success": True, "errors": []},
                {
                    "success": False,
                    "errors": [
                        {
                            "fields": [],
                            "statusCode": "FOO",
                            "message": "Delphic exception message",
                        }
                    ],
                },
            ],
            match=[
                responses.json_params_matcher(
                    {
                        "allOrNone": False,
                        "records": [
                            {
                                "attributes": {"type": "PermissionSetAssignment"},
                                "AssigneeId": "005000000000000",
                                "PermissionSetId": "0PS000000000001",
                            },
                            {
                                "attributes": {"type": "PermissionSetAssignment"},
                                "AssigneeId": "005000000000001",
                                "PermissionSetId": "0PS000000000001",
                            },
                        ],
                    }
                ),
            ],
        )

        with pytest.raises(CumulusCIException):
            task()

        # Check table output
        expected_table_data = [
            ["Success", "ID", "Message"],
            [True, "0Pa000000000000", "-"],
            [False, "-", "Delphic exception message"],
        ]
        table.assert_called_once()
        assert expected_table_data in table.call_args[0]


class TestCreatePermissionSetLicense:
    @responses.activate
    def test_create_permsetlicense(self):
        task = create_task(
            AssignPermissionSetLicenses,
            {
                "api_names": "PermSetLicense1,PermSetLicense2",
            },
        )

        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2C%28SELECT+PermissionSetLicenseId+FROM+PermissionSetLicenseAssignments%29+FROM+User+WHERE+Username+%3D+%27test-cci%40example.com%27",
            status=200,
            json={
                "done": True,
                "totalSize": 1,
                "records": [
                    {
                        "Id": "005000000000000",
                        "PermissionSetLicenseAssignments": {
                            "done": True,
                            "totalSize": 1,
                            "records": [{"PermissionSetLicenseId": "0PL000000000000"}],
                        },
                    }
                ],
            },
        )
        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2CDeveloperName+FROM+PermissionSetLicense+WHERE+DeveloperName+IN+%28%27PermSetLicense1%27%2C+%27PermSetLicense2%27%29",
            status=200,
            json={
                "done": True,
                "totalSize": 1,
                "records": [
                    {
                        "Id": "0PL000000000000",
                        "DeveloperName": "PermSetLicense1",
                    },
                    {
                        "Id": "0PL000000000001",
                        "DeveloperName": "PermSetLicense2",
                    },
                ],
            },
        )
        responses.add(
            method="POST",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/composite/sobjects",
            status=200,
            json=[{"id": "0Pa000000000001", "success": True, "errors": []}],
            match=[
                responses.json_params_matcher(
                    {
                        "allOrNone": False,
                        "records": [
                            {
                                "attributes": {"type": "PermissionSetLicenseAssign"},
                                "AssigneeId": "005000000000000",
                                "PermissionSetLicenseId": "0PL000000000001",
                            }
                        ],
                    }
                ),
            ],
        )

        task()

        assert len(responses.calls) == 3

    @responses.activate
    def test_create_permsetlicense__no_assignments(self):
        task = create_task(
            AssignPermissionSetLicenses,
            {
                "api_names": "PermSetLicense1,PermSetLicense2",
            },
        )

        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2C%28SELECT+PermissionSetLicenseId+FROM+PermissionSetLicenseAssignments%29+FROM+User+WHERE+Username+%3D+%27test-cci%40example.com%27",
            status=200,
            json={
                "done": True,
                "totalSize": 1,
                "records": [
                    {
                        "Id": "005000000000000",
                        # This seems like a bug: the PermissionSetLicenseAssignments sub-query returns None if no PSLs are already assigned instead of returning an "empty list".
                        "PermissionSetLicenseAssignments": None,
                    }
                ],
            },
        )
        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2CDeveloperName+FROM+PermissionSetLicense+WHERE+DeveloperName+IN+%28%27PermSetLicense1%27%2C+%27PermSetLicense2%27%29",
            status=200,
            json={
                "done": True,
                "totalSize": 1,
                "records": [
                    {
                        "Id": "0PL000000000000",
                        "DeveloperName": "PermSetLicense1",
                    },
                    {
                        "Id": "0PL000000000001",
                        "DeveloperName": "PermSetLicense2",
                    },
                ],
            },
        )

        responses.add(
            method="POST",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/composite/sobjects",
            status=200,
            json=[
                {"id": "0Pa000000000000", "success": True, "errors": []},
                {"id": "0Pa000000000001", "success": True, "errors": []},
            ],
            match=[
                responses.json_params_matcher(
                    {
                        "allOrNone": False,
                        "records": [
                            {
                                "attributes": {"type": "PermissionSetLicenseAssign"},
                                "AssigneeId": "005000000000000",
                                "PermissionSetLicenseId": "0PL000000000000",
                            },
                            {
                                "attributes": {"type": "PermissionSetLicenseAssign"},
                                "AssigneeId": "005000000000000",
                                "PermissionSetLicenseId": "0PL000000000001",
                            },
                        ],
                    }
                ),
            ],
        )
        task()

        assert len(responses.calls) == 3

    @responses.activate
    def test_create_permsetlicense__alias(self):
        task = create_task(
            AssignPermissionSetLicenses,
            {
                "api_names": "PermSetLicense1,PermSetLicense2",
                "user_alias": "test",
            },
        )

        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2C%28SELECT+PermissionSetLicenseId+FROM+PermissionSetLicenseAssignments%29+FROM+User+WHERE+Alias+IN+%28%27test%27%29",
            status=200,
            json={
                "done": True,
                "totalSize": 1,
                "records": [
                    {
                        "Id": "005000000000000",
                        "PermissionSetLicenseAssignments": {
                            "done": True,
                            "totalSize": 1,
                            "records": [{"PermissionSetLicenseId": "0PL000000000000"}],
                        },
                    }
                ],
            },
        )
        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2CDeveloperName+FROM+PermissionSetLicense+WHERE+DeveloperName+IN+%28%27PermSetLicense1%27%2C+%27PermSetLicense2%27%29",
            status=200,
            json={
                "done": True,
                "totalSize": 1,
                "records": [
                    {
                        "Id": "0PL000000000000",
                        "DeveloperName": "PermSetLicense1",
                    },
                    {
                        "Id": "0PL000000000001",
                        "DeveloperName": "PermSetLicense2",
                    },
                ],
            },
        )
        responses.add(
            method="POST",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/sobjects/PermissionSetLicenseAssign/",
            status=200,
            json={"id": "0Pa000000000001", "success": True, "errors": []},
        )

        responses.add(
            method="POST",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/composite/sobjects",
            status=200,
            json=[{"id": "0Pa000000000001", "success": True, "errors": []}],
            match=[
                responses.json_params_matcher(
                    {
                        "allOrNone": False,
                        "records": [
                            {
                                "attributes": {"type": "PermissionSetLicenseAssign"},
                                "AssigneeId": "005000000000000",
                                "PermissionSetLicenseId": "0PL000000000001",
                            }
                        ],
                    }
                ),
            ],
        )
        task()

        assert len(responses.calls) == 3

    @responses.activate
    def test_create_permsetlicense__alias_raises(self):
        task = create_task(
            AssignPermissionSetLicenses,
            {
                "api_names": "PermSetLicense1,PermSetLicense2",
                "user_alias": "test",
            },
        )

        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2C%28SELECT+PermissionSetLicenseId+FROM+PermissionSetLicenseAssignments%29+FROM+User+WHERE+Alias+IN+%28%27test%27%29",
            status=200,
            json={
                "done": True,
                "totalSize": 0,
                "records": [],
            },
        )
        with pytest.raises(CumulusCIException):
            task()

    @responses.activate
    def test_create_permsetlicense_raises(self):
        task = create_task(
            AssignPermissionSetLicenses,
            {
                "api_names": "PermSetLicense1,PermSetLicense2,PermSetLicense3",
            },
        )

        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2C%28SELECT+PermissionSetLicenseId+FROM+PermissionSetLicenseAssignments%29+FROM+User+WHERE+Username+%3D+%27test-cci%40example.com%27",
            status=200,
            json={
                "done": True,
                "totalSize": 1,
                "records": [
                    {
                        "Id": "005000000000000",
                        "PermissionSetLicenseAssignments": {
                            "done": True,
                            "totalSize": 1,
                            "records": [{"PermissionSetLicenseId": "0PL000000000000"}],
                        },
                    }
                ],
            },
        )
        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2CDeveloperName+FROM+PermissionSetLicense+WHERE+DeveloperName+IN+%28%27PermSetLicense1%27%2C+%27PermSetLicense2%27%2C+%27PermSetLicense3%27%29",
            status=200,
            json={
                "done": True,
                "totalSize": 1,
                "records": [
                    {
                        "Id": "0PL000000000000",
                        "DeveloperName": "PermSetLicense1",
                    },
                    {
                        "Id": "0PL000000000001",
                        "DeveloperName": "PermSetLicense2",
                    },
                ],
            },
        )

        with pytest.raises(CumulusCIException):
            task()


class TestCreatePermissionSetGroup:
    @responses.activate
    def test_create_permsetgroup(self):
        task = create_task(
            AssignPermissionSetGroups,
            {
                "api_names": "PermSetGroup1,PermSetGroup2",
            },
        )

        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2C%28SELECT+PermissionSetGroupId+FROM+PermissionSetAssignments%29+FROM+User+WHERE+Username+%3D+%27test-cci%40example.com%27",
            status=200,
            json={
                "done": True,
                "totalSize": 1,
                "records": [
                    {
                        "Id": "005000000000000",
                        "PermissionSetAssignments": {
                            "done": True,
                            "totalSize": 1,
                            "records": [{"PermissionSetGroupId": "0PG000000000000"}],
                        },
                    }
                ],
            },
        )
        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2CDeveloperName+FROM+PermissionSetGroup+WHERE+DeveloperName+IN+%28%27PermSetGroup1%27%2C+%27PermSetGroup2%27%29",
            status=200,
            json={
                "done": True,
                "totalSize": 1,
                "records": [
                    {
                        "Id": "0PG000000000000",
                        "DeveloperName": "PermSetGroup1",
                    },
                    {
                        "Id": "0PG000000000001",
                        "DeveloperName": "PermSetGroup2",
                    },
                ],
            },
        )
        responses.add(
            method="POST",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/composite/sobjects",
            status=200,
            json=[{"id": "0Pa000000000001", "success": True, "errors": []}],
            match=[
                responses.json_params_matcher(
                    {
                        "allOrNone": False,
                        "records": [
                            {
                                "attributes": {"type": "PermissionSetAssignment"},
                                "AssigneeId": "005000000000000",
                                "PermissionSetGroupId": "0PG000000000001",
                            }
                        ],
                    }
                ),
            ],
        )

        task()

        assert len(responses.calls) == 3

    @responses.activate
    def test_create_permsetgroup__alias(self):
        task = create_task(
            AssignPermissionSetGroups,
            {
                "api_names": "PermSetGroup1,PermSetGroup2",
                "user_alias": "test",
            },
        )

        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2C%28SELECT+PermissionSetGroupId+FROM+PermissionSetAssignments%29+FROM+User+WHERE+Alias+IN+%28%27test%27%29",
            status=200,
            json={
                "done": True,
                "totalSize": 1,
                "records": [
                    {
                        "Id": "005000000000000",
                        "PermissionSetAssignments": {
                            "done": True,
                            "totalSize": 1,
                            "records": [{"PermissionSetGroupId": "0PG000000000000"}],
                        },
                    }
                ],
            },
        )
        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2CDeveloperName+FROM+PermissionSetGroup+WHERE+DeveloperName+IN+%28%27PermSetGroup1%27%2C+%27PermSetGroup2%27%29",
            status=200,
            json={
                "done": True,
                "totalSize": 1,
                "records": [
                    {
                        "Id": "0PG000000000000",
                        "DeveloperName": "PermSetGroup1",
                    },
                    {
                        "Id": "0PG000000000001",
                        "DeveloperName": "PermSetGroup2",
                    },
                ],
            },
        )
        responses.add(
            method="POST",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/composite/sobjects",
            status=200,
            json=[{"id": "0Pa000000000001", "success": True, "errors": []}],
            match=[
                responses.json_params_matcher(
                    {
                        "allOrNone": False,
                        "records": [
                            {
                                "attributes": {"type": "PermissionSetAssignment"},
                                "AssigneeId": "005000000000000",
                                "PermissionSetGroupId": "0PG000000000001",
                            }
                        ],
                    }
                ),
            ],
        )

        task()

        assert len(responses.calls) == 3

    @responses.activate
    def test_create_permsetgroup__alias_raises(self):
        task = create_task(
            AssignPermissionSetGroups,
            {
                "api_names": "PermSetGroup1,PermSetGroup2",
                "user_alias": "test",
            },
        )

        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2C%28SELECT+PermissionSetGroupId+FROM+PermissionSetAssignments%29+FROM+User+WHERE+Alias+IN+%28%27test%27%29",
            status=200,
            json={
                "done": True,
                "totalSize": 0,
                "records": [],
            },
        )
        with pytest.raises(CumulusCIException):
            task()

    @responses.activate
    def test_create_permsetgroup_raises(self):
        task = create_task(
            AssignPermissionSetGroups,
            {
                "api_names": "PermSetGroup1,PermSetGroup2,PermSetGroup3",
            },
        )

        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2C%28SELECT+PermissionSetGroupId+FROM+PermissionSetAssignments%29+FROM+User+WHERE+Username+%3D+%27test-cci%40example.com%27",
            status=200,
            json={
                "done": True,
                "totalSize": 1,
                "records": [
                    {
                        "Id": "005000000000000",
                        "PermissionSetAssignments": {
                            "done": True,
                            "totalSize": 1,
                            "records": [{"PermissionSetGroupId": "0PG000000000000"}],
                        },
                    }
                ],
            },
        )
        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/query/?q=SELECT+Id%2CDeveloperName+FROM+PermissionSetGroup+WHERE+DeveloperName+IN+%28%27PermSetGroup1%27%2C+%27PermSetGroup2%27%2C+%27PermSetGroup3%27%29",
            status=200,
            json={
                "done": True,
                "totalSize": 1,
                "records": [
                    {
                        "Id": "0PG000000000000",
                        "DeveloperName": "PermSetGroup1",
                    },
                    {
                        "Id": "0PG000000000001",
                        "DeveloperName": "PermSetGroup2",
                    },
                ],
            },
        )

        with pytest.raises(CumulusCIException):
            task()



class classproperty(object):
    """Similar to @property, but allows class-level properties.

    That is, a property whose getter is like a classmethod."""

    def __init__(self, fget):
        self.fget = fget

    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)



import tempfile
import typing as T
from collections import defaultdict
from contextlib import contextmanager
from pathlib import Path
from unittest.mock import MagicMock

from sqlalchemy import Column, MetaData, Table, Unicode, create_engine, func, inspect
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.orm import Session

from cumulusci.core.enums import StrEnum
from cumulusci.core.exceptions import BulkDataException, TaskOptionsError
from cumulusci.core.utils import process_bool_arg
from cumulusci.salesforce_api.org_schema import get_org_schema
from cumulusci.tasks.bulkdata.dates import adjust_relative_dates
from cumulusci.tasks.bulkdata.mapping_parser import (
    CaseInsensitiveDict,
    MappingLookup,
    MappingStep,
    parse_from_yaml,
    validate_and_inject_mapping,
)
from cumulusci.tasks.bulkdata.query_transformers import (
    ID_TABLE_NAME,
    AddLookupsToQuery,
    AddMappingFiltersToQuery,
    AddPersonAccountsToQuery,
    AddRecordTypesToQuery,
)
from cumulusci.tasks.bulkdata.step import (
    DEFAULT_BULK_BATCH_SIZE,
    DataApi,
    DataOperationJobResult,
    DataOperationStatus,
    DataOperationType,
    RestApiDmlOperation,
    get_dml_operation,
)
from cumulusci.tasks.bulkdata.upsert_utils import (
    AddUpsertsToQuery,
    extract_upsert_key_data,
    needs_etl_upsert,
)
from cumulusci.tasks.bulkdata.utils import (
    RowErrorChecker,
    SqlAlchemyMixin,
    sql_bulk_insert_from_records,
)
from cumulusci.tasks.salesforce import BaseSalesforceApiTask


class LoadData(SqlAlchemyMixin, BaseSalesforceApiTask):
    """Perform Bulk API operations to load data defined by a mapping from a local store into an org."""

    task_options = {
        "database_url": {
            "description": "The database url to a database containing the test data to load"
        },
        "mapping": {
            "description": "The path to a yaml file containing mappings of the database fields to Salesforce object fields",
            "required": True,
        },
        "start_step": {
            "description": "If specified, skip steps before this one in the mapping",
            "required": False,
        },
        "sql_path": {
            "description": "If specified, a database will be created from an SQL script at the provided path"
        },
        "ignore_row_errors": {
            "description": "If True, allow the load to continue even if individual rows fail to load."
        },
        "reset_oids": {
            "description": "If True (the default), and the _sf_ids tables exist, reset them before continuing.",
            "required": False,
        },
        "bulk_mode": {
            "description": "Set to Serial to force serial mode on all jobs. Parallel is the default."
        },
        "inject_namespaces": {
            "description": "If True, the package namespace prefix will be "
            "automatically added to (or removed from) objects "
            "and fields based on the name used in the org. Defaults to True."
        },
        "drop_missing_schema": {
            "description": "Set to True to skip any missing objects or fields instead of stopping with an error."
        },
        "set_recently_viewed": {
            "description": "By default, the first 1000 records inserted via the Bulk API will be set as recently viewed. If fewer than 1000 records are inserted, existing objects of the same type being inserted will also be set as recently viewed.",
        },
        "org_shape_match_only": {
            "description": "When True, all path options are ignored and only a dataset matching the org shape name will be loaded. Defaults to False."
        },
        "enable_rollback": {
            "description": "When True, performs rollback operation incase of error. Defaults to False"
        },
    }
    row_warning_limit = 10

    def _init_options(self, kwargs):
        super(LoadData, self)._init_options(kwargs)

        self.options["ignore_row_errors"] = process_bool_arg(
            self.options.get("ignore_row_errors") or False
        )
        self._init_dataset()
        self.reset_oids = self.options.get("reset_oids", True)
        self.bulk_mode = (
            self.options.get("bulk_mode") and self.options.get("bulk_mode").title()
        )
        if self.bulk_mode and self.bulk_mode not in ["Serial", "Parallel"]:
            raise TaskOptionsError("bulk_mode must be either Serial or Parallel")

        inject_namespaces = self.options.get("inject_namespaces")
        self.options["inject_namespaces"] = process_bool_arg(
            True if inject_namespaces is None else inject_namespaces
        )
        self.options["drop_missing_schema"] = process_bool_arg(
            self.options.get("drop_missing_schema") or False
        )
        self.options["set_recently_viewed"] = process_bool_arg(
            self.options.get("set_recently_viewed", True)
        )
        self.options["enable_rollback"] = process_bool_arg(
            self.options.get("enable_rollback", False)
        )
        self._id_generators = {}
        self._old_format = False
        self.ID_TABLE_NAME = ID_TABLE_NAME

    def _init_dataset(self):
        """Find the dataset paths to use with the following sequence:
        1. If `org_shape_match_only` is True (defaults False),
           unset any other path options that may have been supplied.
        2. Prefer a supplied `database_url`.
        3. If `sql_path` was supplied, but not `mapping`, use default `mapping` value.
        4. If `mapping` was supplied, but not `sql_path`, use default `sql_path` value.
        5. If no path options were supplied, look for a dataset matching the org shape.
        6. If no matching dataset was found AND `org_shape_match_only` is False,
           look for a dataset with the default `mapping` and `sql_path` values
           (as previously defaulted in the standard library yml).
        """
        self.options["org_shape_match_only"] = process_bool_arg(
            self.options.get("org_shape_match_only", False)
        )
        if self.options["org_shape_match_only"]:
            self.options["mapping"] = None
            self.options["sql_path"] = None
            self.options["database_url"] = None
            self.logger.warning(
                "The `default_dataset_only` option has been deprecated. "
                "Please switch to the `load_sample_data` task."
            )

        self.options.setdefault("database_url", None)
        if self.options.get("database_url"):
            # prefer database_url if it's set
            self.options["sql_path"] = None
        elif self.options.get("sql_path"):
            self.options.setdefault("mapping", "datasets/mapping.yml")
        elif self.options.get("mapping"):
            self.options.setdefault("sql_path", "datasets/sample.sql")
        elif found_dataset := (
            self._find_matching_dataset() or self._find_default_dataset()
        ):  # didn't get either database_url or sql_path
            mapping_path, dataset_path = found_dataset
            self.options["mapping"] = mapping_path
            self.options["sql_path"] = dataset_path
        else:
            self.has_dataset = False
            return
        self.has_dataset = True

    def _find_matching_dataset(self) -> T.Optional[T.Tuple[str, str]]:
        org_shape = self.org_config.lookup("config_name") if self.org_config else None
        if not org_shape:
            return None  # persistent org
        dataset_folder = f"datasets/{org_shape}"
        if Path(dataset_folder).exists():
            # check for dataset.sql and mapping.yml
            mapping_path = f"{dataset_folder}/{org_shape}.mapping.yml"
            dataset_path = f"{dataset_folder}/{org_shape}.dataset.sql"
            if Path(mapping_path).exists() and Path(dataset_path).exists():
                return (mapping_path, dataset_path)
            else:
                self.logger.warning(
                    f"Found datasets/{org_shape} but it did not contain {org_shape}.mapping.yml and {org_shape}.dataset.yml."
                )
        return None

    def _find_default_dataset(self) -> T.Optional[T.Tuple[str, str]]:
        if self.options["org_shape_match_only"]:
            return None
        dataset_path = "datasets/sample.sql"
        mapping_path = "datasets/mapping.yml"
        if Path(dataset_path).exists() and Path(mapping_path).exists():
            return (mapping_path, dataset_path)
        return None

    def _run_task(self):
        if not self.has_dataset:
            if org_shape := self.org_config.lookup("config_name"):
                self.logger.info(
                    f"No data will be loaded because there was no dataset found matching your org shape name ('{org_shape}')."
                )
            else:
                self.logger.info(
                    "No data will be loaded because this is a persistent org and no dataset was specified."
                )
            return
        self._init_mapping()
        with self._init_db():
            self._expand_mapping()
            self._initialize_id_table(self.reset_oids)
            start_step = self.options.get("start_step")
            started = False
            results = {}
            for name, mapping in self.mapping.items():
                # Skip steps until start_step
                if not started and start_step and name != start_step:
                    self.logger.info(f"Skipping step: {name}")
                    continue

                started = True

                self.logger.info(f"Running step: {name}")
                result = self._execute_step(mapping)
                if result.status is DataOperationStatus.JOB_FAILURE:
                    raise BulkDataException(
                        f"Step {name} did not complete successfully: {','.join(result.job_errors)}"
                    )

                if name in self.after_steps:
                    for after_name, after_step in self.after_steps[name].items():
                        self.logger.info(f"Running post-load step: {after_name}")
                        result = self._execute_step(after_step)
                        if result.status is DataOperationStatus.JOB_FAILURE:
                            raise BulkDataException(
                                f"Step {after_name} did not complete successfully: {','.join(result.job_errors)}"
                            )
                results[name] = StepResultInfo(
                    mapping.sf_object, result, mapping.record_type
                )
        if self.options["set_recently_viewed"]:
            try:
                self.logger.info("Setting records to 'recently viewed'.")
                set_recently_viewed = self._set_viewed()
            except Exception as e:
                self.logger.warning(f"Could not set recently viewed because {e}")
                set_recently_viewed = [SetRecentlyViewedInfo("ALL", e)]
        else:
            set_recently_viewed = False

        self.return_values = {
            "step_results": {
                step_name: result_info.simplify()
                for step_name, result_info in results.items()
            },
        }
        if set_recently_viewed is not False:
            self.return_values["set_recently_viewed"] = set_recently_viewed

    def _execute_step(
        self, mapping: MappingStep
    ) -> T.Union[DataOperationJobResult, MagicMock]:
        """Load data for a single step."""

        if "RecordTypeId" in mapping.fields:
            conn = self.session.connection()
            self._load_record_types([mapping.sf_object], conn)
            self.session.commit()

        step, query = self.configure_step(mapping)

        with tempfile.TemporaryFile(mode="w+t") as local_ids:
            # Store the previous values of the records before upsert
            # This is so that we can perform rollback
            if (
                mapping.action
                in [
                    DataOperationType.ETL_UPSERT,
                    DataOperationType.UPSERT,
                    DataOperationType.UPDATE,
                ]
                and self.options["enable_rollback"]
            ):
                UpdateRollback.prepare_for_rollback(
                    self, step, self._stream_queried_data(mapping, local_ids, query)
                )
            step.start()
            step.load_records(self._stream_queried_data(mapping, local_ids, query))
            step.end()

            # Process Job Results
            if step.job_result.status is not DataOperationStatus.JOB_FAILURE:
                local_ids.seek(0)
                self._process_job_results(mapping, step, local_ids)
            elif (
                step.job_result.status is DataOperationStatus.JOB_FAILURE
                and self.options["enable_rollback"]
            ):
                Rollback._perform_rollback(self)

            return step.job_result

    def configure_step(self, mapping):
        """Create a step appropriate to the action"""
        bulk_mode = mapping.bulk_mode or self.bulk_mode or "Parallel"
        api_options = {"batch_size": mapping.batch_size, "bulk_mode": bulk_mode}

        fields = mapping.get_load_field_list()

        # implement "smart" upsert
        if mapping.action == DataOperationType.SMART_UPSERT:
            if needs_etl_upsert(mapping, self.sf):
                mapping.action = DataOperationType.ETL_UPSERT
            else:
                mapping.action = DataOperationType.UPSERT

        if mapping.action == DataOperationType.ETL_UPSERT:
            extract_upsert_key_data(
                mapping.sf_object,
                mapping.update_key,
                self,
                self.metadata,
                self.session.connection(),
            )

            # If we treat "Id" as an "external_id_name" then it's
            # allowed to be sparse.
            api_options["update_key"] = "Id"
            action = DataOperationType.UPSERT
            fields.append("Id")
        elif mapping.action == DataOperationType.UPSERT:
            self.check_simple_upsert(mapping)
            api_options["update_key"] = mapping.update_key[0]
            action = DataOperationType.UPSERT
        else:
            action = mapping.action

        query = self._query_db(mapping)

        step = get_dml_operation(
            sobject=mapping.sf_object,
            operation=action,
            api_options=api_options,
            context=self,
            fields=fields,
            api=mapping.api,
            volume=query.count(),
        )
        return step, query

    def check_simple_upsert(self, mapping):
        """Check that this upsert is correct."""
        if needs_etl_upsert(mapping, self.sf):
            raise BulkDataException(
                f"This update key is not compatible with a simple upsert: `{','.join(mapping.update_key)}`. "
                "Use `action: ETL_UPSERT` instead."
            )

    def _stream_queried_data(self, mapping, local_ids, query):
        """Get data from the local db"""

        statics = self._get_statics(mapping)
        total_rows = 0

        if mapping.anchor_date:
            date_context = mapping.get_relative_date_context(
                mapping.get_load_field_list(), self.sf
            )
        # Clamping the yield from the query ensures we do not
        # create more Bulk API batches than expected, regardless
        # of batch size, while capping memory usage.
        batch_size = mapping.batch_size or DEFAULT_BULK_BATCH_SIZE
        for row in query.yield_per(batch_size):
            total_rows += 1
            pkey = row[0]
            row = list(row[1:]) + statics

            if mapping.anchor_date and (date_context[0] or date_context[1]):
                row = adjust_relative_dates(
                    mapping, date_context, row, DataOperationType.INSERT
                )

            if mapping.action is DataOperationType.UPDATE:
                if len(row) > 1 and all([f is None for f in row[1:]]):
                    # Skip update rows that contain no values
                    total_rows -= 1
                    continue

            local_ids.write(str(pkey) + "\n")
            yield row

        self.logger.info(
            f"Prepared {total_rows} rows for {mapping.action.value} to {mapping.sf_object}."
        )

    def _load_record_types(self, sobjects, conn):
        """Persist record types for the given sObjects into the database."""
        for sobject in sobjects:
            table_name = sobject + "_rt_target_mapping"
            self._extract_record_types(
                sobject, table_name, conn, self.org_config.is_person_accounts_enabled
            )

    def _get_statics(self, mapping):
        """Return the static values (not column names) to be appended to
        records for this mapping."""
        statics = list(mapping.static.values())
        if mapping.record_type:
            query = (
                f"SELECT Id FROM RecordType WHERE SObjectType='{mapping.sf_object}'"
                f"AND DeveloperName = '{mapping.record_type}' LIMIT 1"
            )
            records = self.sf.query(query)["records"]
            if records:
                record_type_id = records[0]["Id"]
            else:
                raise BulkDataException(f"Cannot find RecordType with query `{query}`")
            statics.append(record_type_id)

        return statics

    def _query_db(self, mapping):
        """Build a query to retrieve data from the local db.

        Includes columns from the mapping
        as well as joining to the id tables to get real SF ids
        for lookups.
        """
        model = self.models[mapping.table]

        id_column = model.__table__.primary_key.columns.keys()[0]
        columns = [getattr(model, id_column)]

        table_cases = CaseInsensitiveDict(model.__table__.columns)
        for name, f in mapping.fields.items():
            if name not in ("Id", "RecordTypeId", "RecordType"):
                column = table_cases.get(f)
                columns.append(column)

        query = self.session.query(*columns)

        classes = [
            AddRecordTypesToQuery,
            AddMappingFiltersToQuery,
            AddUpsertsToQuery,
        ]
        transformers = [
            AddLookupsToQuery(mapping, self.metadata, model, self._old_format)
        ]
        transformers.extend([cls(mapping, self.metadata, model) for cls in classes])

        if mapping.sf_object == "Contact" and self._can_load_person_accounts(mapping):
            transformers.append(AddPersonAccountsToQuery(mapping, self.metadata, model))

        for transformer in transformers:
            query = transformer.add_columns(query)

        for transformer in transformers:
            query = transformer.add_filters(query)

        for transformer in transformers:
            query = transformer.add_outerjoins(query)

        query = self._sort_by_lookups(query, mapping, model)
        return query

    def _sort_by_lookups(self, query, mapping, model):
        lookups = [lookup for lookup in mapping.lookups.values() if not lookup.after]
        for lookup in lookups:
            key_field = lookup.get_lookup_key_field(model)
            lookup_column = getattr(model, key_field)
            query = query.order_by(lookup_column)

        return query

    def _process_job_results(self, mapping, step, local_ids):
        """Get the job results and process the results. If we're raising for
        row-level errors, do so; if we're inserting, store the new Ids."""

        is_insert_or_upsert = mapping.action in (
            DataOperationType.INSERT,
            DataOperationType.UPSERT,
            DataOperationType.ETL_UPSERT,
        )

        conn = self.session.connection()
        sf_id_results = self._generate_results_id_map(step, local_ids)

        for i in range(len(sf_id_results)):
            # Check for old_format of load sql files
            if str(sf_id_results[i][0]).isnumeric():
                self._old_format = True
                # Set id column with new naming format (<sobject> - <counter>)
                sf_id_results[i][0] = mapping.table + "-" + str(sf_id_results[i][0])
            else:
                break
        # If we know we have no successful inserts, don't attempt to persist Ids.
        # Do, however, drain the generator to get error-checking behavior.
        if is_insert_or_upsert and (
            step.job_result.records_processed - step.job_result.total_row_errors
        ):
            table = self.metadata.tables[self.ID_TABLE_NAME]
            sql_bulk_insert_from_records(
                connection=conn,
                table=table,
                columns=("id", "sf_id"),
                record_iterable=sf_id_results,
            )

        # Contact records for Person Accounts are inserted during an Account
        # sf_object step.  Insert records into the Contact ID table for
        # person account Contact records so lookups to
        # person account Contact records get populated downstream as expected.
        if (
            is_insert_or_upsert
            and mapping.sf_object == "Contact"
            and self._can_load_person_accounts(mapping)
        ):
            account_id_lookup = mapping.lookups.get("AccountId")
            if account_id_lookup:
                sql_bulk_insert_from_records(
                    connection=conn,
                    table=self.metadata.tables[self.ID_TABLE_NAME],
                    columns=("id", "sf_id"),
                    record_iterable=self._generate_contact_id_map_for_person_accounts(
                        mapping, account_id_lookup, conn
                    ),
                )

        if is_insert_or_upsert:
            self.session.commit()

    def _generate_results_id_map(self, step, local_ids):
        """Consume results from load and prepare rows for id table.
        Raise BulkDataException on row errors if configured to do so.
        Adds created records into insert_rollback Table
        Performs rollback in case of any errors if enable_rollback is True"""
        error_checker = RowErrorChecker(
            self.logger, self.options["ignore_row_errors"], self.row_warning_limit
        )
        local_ids = (lid.strip("\n") for lid in local_ids)
        sf_id_results = []
        created_results = []
        failed_results = []
        for result, local_id in zip(step.get_results(), local_ids):
            if result.success:
                sf_id_results.append([local_id, result.id])
                if result.created:
                    created_results.append([result.id])
            else:
                failed_results.append([result, local_id])

        # We record failed_results separately since if a unsuccesful record
        # was in between, it would not store all the successful ids
        for result, local_id in failed_results:
            try:
                error_checker.check_for_row_error(result, local_id)
            except Exception as e:
                if self.options["enable_rollback"]:
                    CreateRollback.prepare_for_rollback(self, step, created_results)
                    Rollback._perform_rollback(self)
                raise e
        if self.options["enable_rollback"]:
            CreateRollback.prepare_for_rollback(self, step, created_results)
        return sf_id_results

    def _initialize_id_table(self, should_reset_table):
        """initalize or find table to hold the inserted SF Ids

        The table has a name like xxx_sf_ids and has just two columns, id and sf_id.

        If the table already exists, should_reset_table determines whether to
        drop and recreate it or not.
        """

        already_exists = self.ID_TABLE_NAME in self.metadata.tables

        if already_exists and not should_reset_table:
            return
        elif already_exists:
            self.metadata.remove(self.metadata.tables[self.ID_TABLE_NAME])
        id_table = Table(
            self.ID_TABLE_NAME,
            self.metadata,
            Column("id", Unicode(255), primary_key=True),
            Column("sf_id", Unicode(18)),
        )
        if id_table.exists():
            id_table.drop()
        id_table.create()

    def _sqlite_load(self):
        """Read a SQLite script and initialize the in-memory database."""
        conn = self.session.connection()
        cursor = conn.connection.cursor()
        with open(self.options["sql_path"], "r", encoding="utf-8") as f:
            try:
                cursor.executescript(f.read())
            finally:
                cursor.close()
        # self.session.flush()

    @contextmanager
    def _init_db(self):
        """Initialize the database and automapper."""
        # initialize the DB engine
        with self._database_url() as database_url:
            parent_engine = create_engine(database_url)
            with parent_engine.connect() as connection:
                # initialize the DB session
                self.session = Session(connection)

                if self.options.get("sql_path"):
                    self._sqlite_load()

                # initialize DB metadata
                self.metadata = MetaData()
                self.metadata.bind = connection
                self.inspector = inspect(parent_engine)

                # empty the record of initalized tables
                Rollback._initialized_rollback_tables_api = {}

                # initialize the automap mapping
                self.base = automap_base(bind=connection, metadata=self.metadata)
                self.base.prepare(connection, reflect=True)

                # Loop through mappings and reflect each referenced table
                self.models = {}
                for name, mapping in self.mapping.items():
                    if mapping.table not in self.models:
                        try:
                            self.models[mapping.table] = self.base.classes[
                                mapping.table
                            ]
                        except KeyError as e:
                            raise BulkDataException(f"Table not found in dataset: {e}")

                    # create any Record Type tables we need
                    if "RecordTypeId" in mapping.fields:
                        self._create_record_type_table(
                            mapping.get_destination_record_type_table()
                        )
                self.metadata.create_all()

                self._validate_org_has_person_accounts_enabled_if_person_account_data_exists()
                yield

    def _init_mapping(self):
        """Load a YAML mapping file."""
        mapping_file_path = self.options.get("mapping")
        if not mapping_file_path:
            raise TaskOptionsError("Mapping file path required")

        self.mapping = parse_from_yaml(mapping_file_path)

        validate_and_inject_mapping(
            mapping=self.mapping,
            sf=self.sf,
            namespace=self.project_config.project__package__namespace,
            data_operation=DataOperationType.INSERT,
            inject_namespaces=self.options["inject_namespaces"],
            drop_missing=self.options["drop_missing_schema"],
        )

    def _expand_mapping(self):
        """Walk the mapping and generate any required 'after' steps
        to handle dependent and self-lookups."""
        # Expand the mapping to handle dependent lookups
        self.after_steps = defaultdict(dict)

        for step in self.mapping.values():
            if any([lookup.after for lookup in step.lookups.values()]):
                # We have deferred/dependent lookups.
                # Synthesize mapping steps for them.

                sobject = step.sf_object
                after_list = {
                    lookup.after for lookup in step.lookups.values() if lookup.after
                }

                for after in after_list:
                    lookups = {
                        lookup_field: lookup
                        for lookup_field, lookup in step.lookups.items()
                        if lookup.after == after
                    }
                    name = f"Update {sobject} Dependencies After {after}"
                    mapping = MappingStep(
                        sf_object=sobject,
                        api=step.api,
                        action="update",
                        table=step.table,
                    )
                    mapping.lookups["Id"] = MappingLookup(
                        name="Id",
                        table=step["table"],
                        key_field=self.models[
                            step["table"]
                        ].__table__.primary_key.columns.keys()[0],
                    )
                    for lookup in lookups:
                        mapping.lookups[lookup] = lookups[lookup].copy()
                        mapping.lookups[lookup].after = None

                    self.after_steps[after][name] = mapping

    def _validate_org_has_person_accounts_enabled_if_person_account_data_exists(self):
        """
        To ensure data is loaded from the dataset as expected as well as avoid partial
        failues, raise a BulkDataException if there exists Account or Contact records with
        IsPersonAccount as 'true' but the org does not have person accounts enabled.
        """
        for mapping in self.mapping.values():
            if mapping.sf_object in [
                "Account",
                "Contact",
            ] and self._db_has_person_accounts_column(mapping):
                table = self.models[mapping.table].__table__
                if (
                    self.session.query(table)
                    .filter(table.columns.get("IsPersonAccount") == "true")
                    .first()
                    and not self.org_config.is_person_accounts_enabled
                ):
                    raise BulkDataException(
                        "Your dataset contains Person Account data but Person Accounts is not enabled for your org."
                    )

    def _db_has_person_accounts_column(self, mapping):
        """Returns whether "IsPersonAccount" is a column in mapping's table."""
        return (
            self.models[mapping.table].__table__.columns.get("IsPersonAccount")
            is not None
        )

    def _can_load_person_accounts(self, mapping) -> bool:
        """Returns whether person accounts can be loaded:
        - The mapping has a "IsPersonAccount" column
        - Person Accounts is enabled in the org.
        """
        return (
            self._db_has_person_accounts_column(mapping)
            and self.org_config.is_person_accounts_enabled
        )

    def _generate_contact_id_map_for_person_accounts(
        self, contact_mapping, account_id_lookup, conn
    ):
        """
        Yields (local_id, sf_id) for Contact records where IsPersonAccount
        is true that can handle large data volumes.

        We know a Person Account record is related to one and only one Contact
        record.  Therefore, we can map local Contact IDs to Salesforce IDs
        by previously inserted Account records:
        - Query the DB to get the map: Salesforce Account ID ->
          local Contact ID
        - Query Salesforce to get the map: Salesforce Account ID ->
          Salesforce Contact ID
        - Merge the maps
        """
        # Contact table columns
        contact_model = self.models[contact_mapping.table]

        contact_id_column = getattr(
            contact_model, contact_model.__table__.primary_key.columns.keys()[0]
        )
        account_id_column = getattr(
            contact_model, account_id_lookup.get_lookup_key_field(contact_model)
        )

        # Account ID table + column
        account_sf_ids_table = account_id_lookup.aliased_table
        account_sf_id_column = account_sf_ids_table.columns["sf_id"]

        # Query the Contact table for person account contact records so we can
        # create a Map: Account SF ID --> Contact ID.  Outer join the
        # Account SF IDs table to get each Contact's associated
        # Account SF ID.
        if self._old_format:
            query = (
                self.session.query(contact_id_column, account_sf_id_column)
                .filter(
                    func.lower(contact_model.__table__.columns.get("IsPersonAccount"))
                    == "true"
                )
                .outerjoin(
                    account_sf_ids_table,
                    account_sf_ids_table.columns["id"]
                    == str(account_id_lookup.table) + "-" + account_id_column,
                )
            )
        else:
            query = (
                self.session.query(contact_id_column, account_sf_id_column)
                .filter(
                    func.lower(contact_model.__table__.columns.get("IsPersonAccount"))
                    == "true"
                )
                .outerjoin(
                    account_sf_ids_table,
                    account_sf_ids_table.columns["id"] == account_id_column,
                )
            )

        # Stream the results so we can process batches of 200 Contacts
        # in case we have large data volumes.
        query_result = conn.execution_options(stream_results=True).execute(
            query.statement
        )

        while True:
            # While we have a chunk to process
            chunk = query_result.fetchmany(200)
            if not chunk:
                break

            # Collect Map: Account SF ID --> Contact ID
            contact_ids_by_account_sf_id = {record[1]: record[0] for record in chunk}

            # Query Map: Account SF ID --> Contact SF ID
            # It's safe to use query_all since the chunk size to 200.
            for record in self.sf.query_all(
                "SELECT Id, AccountId FROM Contact WHERE IsPersonAccount = true AND AccountId IN ('{}')".format(
                    "','".join(contact_ids_by_account_sf_id.keys())
                )
            )["records"]:
                contact_id = contact_ids_by_account_sf_id.get(record["AccountId"])
                contact_sf_id = record["Id"]

                # Join maps together to get tuple (Contact ID, Contact SF ID) to insert into step's ID Table.
                if self._old_format:
                    yield (contact_mapping.table + "-" + str(contact_id), contact_sf_id)
                else:
                    yield (contact_id, contact_sf_id)

    def _set_viewed(self) -> T.List["SetRecentlyViewedInfo"]:
        """Set items as recently viewed. Filter out custom objects without custom tabs."""
        object_names = set()
        custom_objects = set()
        results = []

        # Separate standard and custom objects
        for mapping in self.mapping.values():
            object_name = mapping.sf_object
            if object_name.endswith("__c"):
                custom_objects.add(object_name)
            else:
                object_names.add(object_name)
        # collect SobjectName that have custom tabs
        if custom_objects:
            try:
                custom_tab_objects = self.sf.query_all(
                    "SELECT SObjectName FROM TabDefinition WHERE IsCustom = true AND SObjectName IN ('{}')".format(
                        "','".join(sorted(custom_objects))
                    )
                )
                for record in custom_tab_objects["records"]:
                    object_names.add(record["SobjectName"])
            except Exception as e:
                self.logger.warning(
                    f"Cannot get the list of custom tabs to set recently viewed status on them. Error: {e}"
                )
        with get_org_schema(
            self.sf, self.org_config, included_objects=object_names, force_recache=True
        ) as org_schema:
            for mapped_item in sorted(object_names):
                if org_schema[mapped_item].mruEnabled:
                    try:
                        self.sf.query_all(
                            f"SELECT Id FROM {mapped_item} ORDER BY CreatedDate DESC LIMIT 1000 FOR VIEW"
                        )
                        results.append(SetRecentlyViewedInfo(mapped_item, None))
                    except Exception as e:
                        self.logger.warning(
                            f"Cannot set recently viewed status for {mapped_item}. Error: {e}"
                        )
                        results.append(SetRecentlyViewedInfo(mapped_item, e))
        return results


class RollbackType(StrEnum):
    """Enum to specify type of rollback"""

    UPSERT = "upsert_rollback"
    INSERT = "insert_rollback"


class Rollback:
    # Store the table name and it's corresponding API (rest or bulk)
    _initialized_rollback_tables_api = {}

    @staticmethod
    def _create_tables_for_rollback(context, step, rollback_type: RollbackType) -> str:
        """Create the tables required for upsert and insert rollback"""
        table_name = f"{step.sobject}_{rollback_type}"

        if table_name not in Rollback._initialized_rollback_tables_api:
            common_columns = [Column("Id", Unicode(255), primary_key=True)]

            additional_columns = (
                [Column(field, Unicode(255)) for field in step.fields if field != "Id"]
                if rollback_type is RollbackType.UPSERT
                else []
            )

            columns = common_columns + additional_columns

            # Create the table
            rollback_table = Table(table_name, context.metadata, *columns)
            rollback_table.create()

            # Store the API in the initialized tables dictionary
            if isinstance(step, RestApiDmlOperation):
                Rollback._initialized_rollback_tables_api[table_name] = DataApi.REST
            else:
                Rollback._initialized_rollback_tables_api[table_name] = DataApi.BULK

        return table_name

    @staticmethod
    def _perform_rollback(context):
        """Perform total rollback"""
        context.logger.info("--Initiated Rollback Procedure--")
        for table in reversed(context.metadata.sorted_tables):
            if table.name.endswith(RollbackType.INSERT):
                CreateRollback._perform_rollback(context, table)
            elif table.name.endswith(RollbackType.UPSERT):
                UpdateRollback._perform_rollback(context, table)
        context.logger.info("--Finished Rollback Procedure--")


class UpdateRollback:
    @staticmethod
    def prepare_for_rollback(context, step, records):
        """Retrieve previous values for records being updated"""
        results, columns = step.get_prev_record_values(records)
        if results:
            table_name = Rollback._create_tables_for_rollback(
                context, step, RollbackType.UPSERT
            )
            conn = context.session.connection()
            sql_bulk_insert_from_records(
                connection=conn,
                table=context.metadata.tables[table_name],
                columns=columns,
                record_iterable=results,
            )

    @staticmethod
    def _perform_rollback(context, table: Table) -> None:
        """Perform rollback for updated records"""
        sf_object = table.name.split(f"_{RollbackType.UPSERT.value}")[0]
        records = context.session.query(table).all()

        if records:
            context.logger.info(f"Reverting upserts for {sf_object}")
            api_options = {"update_key": "Id"}

            # Use get_dml_operation to create an UPSERT step
            step = get_dml_operation(
                sobject=sf_object,
                operation=DataOperationType.UPSERT,
                api_options=api_options,
                context=context,
                fields=[column.name for column in table.columns],
                api=Rollback._initialized_rollback_tables_api[table.name],
                volume=len(records),
            )
            step.start()
            step.load_records(records)
            step.end()
            context.logger.info("Done")


class CreateRollback:
    @staticmethod
    def prepare_for_rollback(context, step, records):
        """Store the sf_ids of all records that were created
        to prepare for rollback"""
        if records:
            table_name = Rollback._create_tables_for_rollback(
                context, step, RollbackType.INSERT
            )
            conn = context.session.connection()
            sql_bulk_insert_from_records(
                connection=conn,
                table=context.metadata.tables[table_name],
                columns=["Id"],
                record_iterable=records,
            )

    @staticmethod
    def _perform_rollback(context, table: Table) -> None:
        """Perform rollback for insert operation"""
        sf_object = table.name.split(f"_{RollbackType.INSERT.value}")[0]
        records = context.session.query(table).all()

        if records:
            context.logger.info(f"Deleting {sf_object} records")
            # Perform DELETE operation using get_dml_operation
            step = get_dml_operation(
                sobject=sf_object,
                operation=DataOperationType.DELETE,
                fields=["Id"],
                api_options={},
                context=context,
                api=Rollback._initialized_rollback_tables_api[table.name],
                volume=len(records),
            )
            step.start()
            step.load_records(records)
            step.end()
            context.logger.info("Done")


class StepResultInfo(T.NamedTuple):
    """Represent a Step Result in a form easily convertible to JSON"""

    sobject: str
    result: DataOperationJobResult
    record_type: str = None

    def simplify(self):
        return {
            "sobject": self.sobject,
            "record_type": self.record_type,
            **self.result.simplify(),
        }


class SetRecentlyViewedInfo(T.NamedTuple):
    """Did the set recently succeed or fail?"""

    sobject: str
    error: T.Optional[Exception]



import cmd
import os
import pdb
import re
import signal
import sys
import textwrap

from robot.libraries.BuiltIn import BuiltIn

from cumulusci.cli.ui import CliTable

# this is code we use to inject a style into the DOM
initialize_highlight_js = """
    if (!window.rdbInitialized) {
        console.log("initializing rdb...");
        window.rdbInitialized = true;
        var new_style = document.createElement('style');
        new_style.type = 'text/css';
        new_style.innerHTML = '.rdbHighlight {box-shadow: 0px 1px 4px 2px inset #FFFF00}';
        document.getElementsByTagName('head')[0].appendChild(new_style);
    };
"""


class DebuggerCli(cmd.Cmd, object):
    intro = textwrap.dedent(
        """
        :::
        ::: Welcome to rdb, the Robot Framework debugger
        :::

        Type help or ? to list commands.
    """
    )
    prompt = "rdb> "

    # Robot redirects sys.stdout, so use the original handle
    # or whatever is passed in.
    def __init__(self, listener, stdout=sys.__stdout__):

        cmd.Cmd.__init__(self, stdout=stdout)

        self.listener = listener
        self.builtin = BuiltIn()

        # some simple aliases.
        self.do_c = self.do_continue
        self.do_le = self.do_locate_elements
        self.do_s = self.do_step

    @property
    def webbrowser(self):
        # This is a property because we can't initialize it until we
        # need it. We'll reinitialize it every time it's used. It's
        # fast enough, and this way it should work even if the user
        # is testing a suite with some playwright and some selenium
        # tests
        libraries = self.builtin.get_library_instance(all=True)
        if "SeleniumLibrary" in libraries:
            return SeleniumProxy(libraries["SeleniumLibrary"])
        elif "Browser" in libraries:
            return BrowserProxy(libraries["Browser"])
        return None

    def default(self, line):
        """Ignore lines that begin with #"""

        if line.startswith("#"):
            return

        elif re.match(r"^[$@&]\{.*\}$", line):
            try:
                value = self.builtin.get_variable_value(line)
                print(value, file=self.stdout)
            except Exception:
                print("unknown variable '{}'".format(line), file=self.stdout)

        else:
            super(DebuggerCli, self).default(line)

    def do_continue(self, arg=None):
        """Let the test continue"""
        return True

    def do_locate_elements(self, locator):
        """Find and highlight all elements that match the given locator

        Example:

            rdb> locate_elements  //button[@title='Learn More']
        """
        try:
            self.webbrowser.highlight_elements(locator)
        except Exception as e:
            print(f"Unable to highlight elements: {e}")

    def do_pdb(self, arg=None):
        """Start pdb

        The context will be this function, which won't be particularly
        useful. This is mostly for debugging the debug code. How meta!
        """

        pdb.Pdb(stdout=self.stdout).set_trace()

    def do_quit(self, arg=None):
        """Cause robot to quit gracefully (all remaining tests will be skipped)"""
        os.kill(os.getpid(), signal.SIGTERM)
        return True

    def do_reset_elements(self, arg=None):
        """Remove all highlighting added by `locate_elements`"""
        try:
            self.webbrowser.restore_element_style()
        except Exception as e:
            print(f"Unable to reset element styles: {e}")

    def do_shell(self, arg):
        """
        Execute a robot framework keyword. (shortcut: !)

        The statement should be formatted just like in a .robot
        file, with two spaces between each argument.

        Example:

            rdb> shell log  hello, world
            -or-
            rdb> ! log  hello, world

        Variables can be specified the same as in a robot file. For
        example, this will set a test variable named ${location}:

            rdb> shell ${location}  get location
        """

        # Split into words with multiple spaces as a separator, then
        # pull out any leading varibles before running the command.
        words = re.split(" {2,}", arg)
        vars = []
        for word in words:
            if re.match(r"[\$@&]\{.*?\}\s*=?", word):
                vars.append(word.rstrip("="))
            else:
                break
        statement = words[len(vars) :]

        try:
            status, result = self.builtin.run_keyword_and_ignore_error(*statement)
            print("status: {}".format(status), file=self.stdout)
            if not vars or status == "FAIL":
                print("result: {}".format(result), file=self.stdout)

            else:
                # Assign test variables given on the command line
                # I don't think this is exactly how robot does it,
                # but I think it's good enough for all normal cases.
                if len(vars) == 1:
                    self.builtin.set_test_variable(vars[0], result)
                    print("{} was set to {}".format(vars[0], result), file=self.stdout)
                else:
                    for (varname, value) in zip(vars, result):
                        self.builtin.set_test_variable(varname, value)
                        print(
                            "{} was set to {}".format(varname, value), file=self.stdout
                        )

        except Exception as e:
            print("error running keyword: {}".format(e), file=self.stdout)

    def do_step(self, arg=None):
        """Run the next step in the test

        This will run the line where the test is currently paused, and then
        stop at the next keyword in the same test case.
        """
        self.listener.do_step()
        return True

    def do_vars(self, arg=None):
        """Print the value of all known variables"""
        vars = self.builtin.get_variables()
        vars = [["Variable", "Value"]] + [list(x) for x in sorted(vars.items())]
        CliTable(vars).echo()

    def do_where(self, arg=None):
        """
        Print the current robot stack (hierarchy of suites and tests,
        ending with the current keyword
        """

        prefix = "  "
        for i, x in enumerate(self.listener.stack):
            indent = prefix * i
            print("{}: {}-> {}".format(i, indent, x.longname), file=self.stdout)
        print("", file=self.stdout)


class SeleniumProxy:
    """Proxy for performing debug operations in a Selenium browser"""

    def __init__(self, library_instance):
        self.selenium = library_instance

    def highlight_elements(self, locator):
        """Highlights Selenium Webdriver elements that match a locator"""
        self.selenium.driver.execute_script(initialize_highlight_js)

        elements = self.selenium.get_webelements(locator)
        for element in elements:
            self.selenium.driver.execute_script(
                "arguments[0].classList.add('rdbHighlight');", element
            )
        print(f"{len(elements)} elements found")

    def restore_element_style(self):
        """Remove the style added by `highlight_elements`"""
        for element in self.selenium.get_webelements("css:.rdbHighlight"):
            self.selenium.driver.execute_script(
                "arguments[0].classList.remove('rdbHighlight')", element
            )


class BrowserProxy:
    """Proxy for performing debug operations in a Playwright browser"""

    def __init__(self, library_instance):
        self.browser = library_instance

    def highlight_elements(self, selector):
        """Highlight one or more elements by applying a custom css class"""
        init_script = f"() => {{{initialize_highlight_js}}};"
        self.browser.evaluate_javascript(None, init_script)

        elements = self.browser.get_elements(selector)
        if elements:
            self.browser.evaluate_javascript(
                selector,
                """(elements) => {
                    for (element of elements) {
                        element.classList.add('rdbHighlight')
                    }
                }""",
                all_elements=True,
            )
        print(f"{len(elements)} elements found")

    def restore_element_style(self):
        """Remove the style added by `highlight_elements`"""
        self.browser.evaluate_javascript(
            ".rdbHighlight",
            """(elements) => {
                for (element of elements) {
                    element.classList.remove('rdbHighlight')
                }
            }""",
            all_elements=True,
        )



from difflib import get_close_matches
from typing import Any, Dict, List

from cumulusci.core.config import BaseConfig, FlowConfig, TaskConfig
from cumulusci.core.exceptions import (
    CumulusCIException,
    FlowNotFoundError,
    TaskNotFoundError,
)


def list_infos(infos: dict) -> List[Dict[str, str]]:
    rv = []
    for info_name, info in infos.items():
        if not info:
            info = {}
        rv.append(
            {
                "name": info_name,
                "description": info.get("description", ""),
                "group": info.get("group"),
            }
        )

    return rv


class BaseTaskFlowConfig(BaseConfig):
    """Base class for all configs that contain tasks and flows"""

    tasks: dict
    flows: dict

    def list_tasks(self) -> List[Dict[str, str]]:
        """Returns a list of task info dictionaries with keys 'name' and 'description'"""
        return list_infos(self.tasks)

    def get_task(self, name: str) -> TaskConfig:
        """Returns a TaskConfig"""
        config = self.lookup(f"tasks__{name}")
        if not config and name not in self.tasks:
            # task does not exist
            error_msg = f"Task not found: {name}"
            suggestion = self.get_suggested_name(name, self.tasks)
            if self.source and self.source.location:
                suggestion += "\n"
                suggestion += "Source of the problem:"
                suggestion += str(self.source.location)
            raise TaskNotFoundError(error_msg + suggestion)
        elif not config:
            # task exists but there is no config at all
            error_msg = f"No configuration found for task: {name}"
            raise CumulusCIException(error_msg)
        elif "class_path" not in config:
            # task exists and there is a config but it has no class_path defined and it is not a base task override
            error_msg = f"Task has no class_path defined: {name}"
            raise CumulusCIException(error_msg)

        return TaskConfig(config)

    def list_flows(self) -> List[Dict[str, str]]:
        """Returns a list of flow info dictionaries with keys 'name' and 'description'"""
        return list_infos(self.flows)

    def get_flow(self, name: str) -> FlowConfig:
        """Returns a FlowConfig"""
        config = self.lookup(f"flows__{name}")
        if not config:
            error_msg = f"Flow not found: {name}"
            suggestion = self.get_suggested_name(name, self.flows)
            raise FlowNotFoundError(error_msg + suggestion)
        return FlowConfig(config)

    def get_suggested_name(self, name: str, steps: Dict[str, Any]) -> str:
        """
        Given a name that cannot be resolved and a list of tasks/flow dicts, returns the nearest match.
        """
        match_list = get_close_matches(name, steps.keys(), n=1)
        if match_list:
            return f'. Did you mean "{match_list[0]}"?'
        else:
            return ""



from contextlib import contextmanager
from typing import Dict, List

import pytest

from cumulusci.tasks.bulkdata.delete import DeleteData


@pytest.fixture()
def ensure_records(create_task, run_code_without_recording, sf):
    """Delete all records of an sobject and create a certain number of new ones"""

    @contextmanager
    def _ensure_records(sobjects: Dict[str, List[Dict[str, str]]]):
        def delete():
            task = create_task(DeleteData, {"objects": ",".join(sobjects.keys())})
            task()

        def setup():
            for obj, records in sobjects.items():
                proxy = getattr(sf, obj)
                for record in records:
                    proxy.create(record)

        run_code_without_recording(delete)
        run_code_without_recording(setup)
        yield
        run_code_without_recording(delete)

    return _ensure_records


@pytest.fixture()
def ensure_accounts(ensure_records):
    """Delete all accounts and create a certain number of new ones"""

    @contextmanager
    def _ensure_accounts(number_of_accounts):
        with ensure_records(
            {
                "Entitlement": [],
                "Account": [
                    {"name": f"Account {i}"} for i in range(number_of_accounts)
                ],
            }
        ):
            yield

    return _ensure_accounts



from datetime import datetime

from dateutil.parser import ParserError
from dateutil.parser import parse as parse_date
from robot.libraries.BuiltIn import BuiltIn
from robot.utils import timestr_to_secs

from cumulusci.robotframework.base_library import BaseLibrary


class Performance(BaseLibrary):
    """
    Keywords for performance testing.

    For more information on how to use these keywords, see
    the "Performance Testing" section of the CumulusCI documentation.
    """

    def elapsed_time_for_last_record(
        self, obj_name, start_field, end_field, order_by, **kwargs
    ):
        """For records representing jobs or processes, compare the record's start-time to its end-time to see how long a process took.

        Arguments:
            obj_name:   SObject to look for last record
            start_field: Name of the datetime field that represents the process start
            end_field: Name of the datetime field that represents the process end
            order_by: Field name to order by. Should be a datetime field, and usually is just the same as end_field.
            where:  Optional Where-clause to use for filtering
            Other keywords are used for filtering as in the Salesforce Query keywordf

        The last matching record queried and summarized.

        Example:

            | ${time_in_seconds} =    Elapsed Time For Last Record
            | ...             obj_name=AsyncApexJob
            | ...             where=ApexClass.Name='BlahBlah'
            | ...             start_field=CreatedDate
            | ...             end_field=CompletedDate
            | ...             order_by=CompletedDate
        """
        if len(order_by.split()) != 1:
            raise Exception("order_by should be a simple field name")
        query = self.salesforce_api._soql_query_builder(
            obj_name,
            select=f"{start_field}, {end_field}",
            order_by=order_by + " DESC NULLS LAST",
            limit=1,
            **kwargs,
        )
        response = self.salesforce_api.soql_query(query)
        results = response["records"]

        if results:
            record = results[0]
            return _duration(record[start_field], record[end_field], record)
        else:
            raise Exception(f"Matching record not found: {query}")

    def start_performance_timer(self):
        """Start an elapsed time stopwatch for performance tests.

        See the docummentation for **Stop Performance Timer** for more
        information.

        Example:

        | Start Performance Timer
        | Do Something
        | Stop Performance Timer
        """
        BuiltIn().set_test_variable("${__start_time}", datetime.now())

    def stop_performance_timer(self):
        """Record the results of a stopwatch. For perf testing.

        This keyword uses Set Test Elapsed Time internally and therefore
        outputs in all of the ways described there.

        Example:

        | Start Performance Timer
        | Do Something
        | Stop Performance Timer

        """
        start_time = self.builtin.get_variable_value("${__start_time}")
        if start_time:
            seconds = (datetime.now() - start_time).seconds
            assert seconds is not None
            self.set_test_elapsed_time(seconds)
        else:
            raise Exception(
                "Elapsed time clock was not started. "
                "Use the Start Elapsed Time keyword to do so."
            )

    def set_test_elapsed_time(self, elapsedtime):
        """This keyword captures a computed rather than measured elapsed time for performance tests.

        For example, if you were performance testing a Salesforce batch process, you might want to
        store the Salesforce-measured elapsed time of the batch process instead of the time measured
        in the CCI client process.

        The keyword takes a single argument which is either a number of seconds or a Robot time string
        (https://robotframework.org/robotframework/latest/libraries/DateTime.html#Time%20formats).

        Using this keyword will automatically add the tag cci_metric_elapsed_time to the test case
        and ${cci_metric_elapsed_time} to the test's variables. cci_metric_elapsed_time is not
        included in Robot's html statistical roll-ups.

        Example:

        | Set Test Elapsed Time       11655.9

        Performance test times are output in the CCI logs and are captured in MetaCI instead of the
        "total elapsed time" measured by Robot Framework. The Robot "test message" is also updated."""

        try:
            seconds = float(elapsedtime)
        except ValueError:
            seconds = timestr_to_secs(elapsedtime)
        assert seconds is not None

        self.builtin.set_test_message(f"Elapsed time set by test : {seconds}")
        self.builtin.set_tags("cci_metric_elapsed_time")
        self.builtin.set_test_variable("${cci_metric_elapsed_time}", seconds)

    def set_test_metric(self, metric: str, value=None):
        """This keyword captures any metric for performance monitoring.

        For example: number of queries, rows processed, CPU usage, etc.

        The keyword takes a metric name, which can be any string, and a value, which
        can be any number.

        Using this keyword will automatically add the tag cci_metric to the test case
        and ${cci_metric_<metric_name>} to the test's variables. These permit downstream
        processing in tools like CCI and MetaCI.

        cci_metric is not included in Robot's html statistical roll-ups.

        Example:

        | Set Test Metric    Max_CPU_Percent    30

        Performance test metrics are output in the CCI logs, log.html and output.xml.
        MetaCI captures them but does not currently have a user interface for displaying
        them."""

        value = float(value)

        self.builtin.set_tags("cci_metric")
        self.builtin.set_test_variable("${cci_metric_%s}" % metric, value)


def _duration(start_date: str, end_date: str, record: dict):
    try:
        start_date = parse_date(start_date)
        end_date = parse_date(end_date)
    except (ParserError, TypeError) as e:
        raise Exception(f"Date parse error: {e} in record {record}")
    duration = end_date - start_date
    return duration.total_seconds()



from unittest import mock

import pytest
from SeleniumLibrary.errors import ElementNotFound

from cumulusci.robotframework.Salesforce import Salesforce


# _init_locators has a special code block
class TestSeleniumLibrary:
    def test_init_locators(self):
        """Verify that locators are initialized if not passed in"""
        with mock.patch.object(Salesforce, "_init_locators"):
            # _init_locators should NOT be called if we pass them in
            sflib = Salesforce(locators={"body": "//whatever"})
            assert not sflib._init_locators.called

            # _init_locators SHOULD be called if we don't pass them in
            sflib = Salesforce()
            sflib._init_locators.assert_called_once()


@mock.patch("robot.libraries.BuiltIn.BuiltIn._get_context")
class TestKeyword_wait_until_salesforce_is_ready:
    @classmethod
    def setup_class(cls):
        cls.sflib = Salesforce(locators={"body": "//whatever"})

    def test_successful_page_load(self, mock_robot_context):
        """Verify that a succesful page load returns no errors"""
        with mock.patch.object(Salesforce, "wait_for_aura", return_value=True):
            self.sflib.wait_until_salesforce_is_ready(timeout="1")

            self.sflib.wait_for_aura.assert_called_once()
            self.sflib.selenium.get_webelement.assert_called_once_with("//whatever")

    def test_reload_on_initial_failure(self, mock_robot_context):
        """Verify that we attempt a reload when we don't find the lightning component"""
        with mock.patch.object(Salesforce, "wait_for_aura", return_value=True):
            with mock.patch.object(
                Salesforce, "_check_for_classic", return_value=False
            ):
                with mock.patch.object(
                    Salesforce, "_check_for_login_failure", return_value=False
                ):
                    with mock.patch.object(
                        self.sflib.selenium,
                        "get_webelement",
                        side_effect=(ElementNotFound(), True),
                    ):
                        self.sflib.wait_until_salesforce_is_ready(timeout="10")
                        self.sflib.selenium.go_to.assert_called_once()

    def test_exception_and_screenshot_on_timeout(self, mock_robot_context):
        """Verify that we throw an appropriate exception after the timeout"""
        with mock.patch.object(Salesforce, "wait_for_aura", return_value=True):
            self.sflib.selenium.get_webelement.side_effect = ElementNotFound()

            with pytest.raises(
                Exception, match="Timed out waiting for a lightning page"
            ):
                # The timeout needs to be longer than the duration of
                # one loop iteration, but less than the retry interval
                # of 5 seconds. Making it longer should still pass the
                # test, it just makes the test run longer than necessary.
                self.sflib.wait_until_salesforce_is_ready(timeout=0.1)

            self.sflib.selenium.capture_page_screenshot.assert_called()


@mock.patch("robot.libraries.BuiltIn.BuiltIn._get_context")
class TestKeyword_breakpoint:
    @classmethod
    def setup_class(cls):
        cls.sflib = Salesforce(locators={"body": "//whatever"})

    def test_breakpoint(self, mock_robot_context):
        """Verify that the keyword doesn't raise an exception"""
        assert self.sflib.breakpoint() is None



import hashlib
from pathlib import Path
from unittest import mock

import pytest
from github3.git import Tree
from github3.repos import Repository

from cumulusci.core.exceptions import GithubException
from cumulusci.tasks.github.util import CommitDir
from cumulusci.utils import temporary_dir


@pytest.fixture
def repo():
    repo = mock.Mock(spec=Repository)
    repo.owner = "SalesforceFoundation"
    repo.name = "TestRepo"
    return repo


class TestCommitDir:
    def test_call(self, repo, tmp_path):
        repo.git_commit.tree.sha = mock.Mock()
        repo.create_blob.return_value = "mocked_sha_hash"
        repo.tree = mock.Mock()
        repo.tree.return_value = Tree(
            {
                "url": "string",
                "sha": "tree-ish-hash",
                "tree": [
                    {
                        "type": "tree",
                        "mode": "100644",
                        "path": "dir",
                        "sha": "bogus1",
                    },
                    {
                        "type": "blob",
                        "mode": "100644",
                        "path": "file_outside_dir",
                        "sha": "bogus2",
                    },
                    {
                        "type": "blob",
                        "mode": "100644",
                        "path": "dir/deleteme",
                        "sha": "deletedfilesha",
                    },
                    {
                        "type": "blob",
                        "mode": "100644",
                        "path": "dir/unchanged",
                        "sha": hashlib.sha1(b"blob 0\0").hexdigest(),
                    },
                    {
                        "type": "blob",
                        "mode": "100644",
                        "path": "dir/modified",
                        "sha": "bogus3",
                    },
                    {
                        "type": "blob",
                        "mode": "100644",
                        "path": "dir/modified_bin",
                        "sha": "bogus4",
                    },
                ],
            },
            None,
        )

        commit = CommitDir(repo)
        Path(tmp_path, "unchanged").write_text("")
        Path(tmp_path, "modified").write_text("modified")
        Path(tmp_path, "new").write_text("new")
        Path(tmp_path, ".hidden").write_text(".hidden")
        Path(tmp_path, "modified_bin").write_bytes(b"\x9c")
        Path(tmp_path, "new_bin").write_bytes(b"\x9c")

        commit(tmp_path, "main", "dir", dry_run=True)
        expected_tree = [
            {"type": "blob", "mode": "100644", "path": "dir/deleteme", "sha": None},
            {
                "type": "blob",
                "mode": "100644",
                "path": "dir/modified",
                "content": "modified",
            },
            {
                "type": "blob",
                "mode": "100644",
                "path": "dir/modified_bin",
                "sha": "mocked_sha_hash",
            },
            {"path": "dir/new", "mode": "100644", "type": "blob", "content": "new"},
            {
                "path": "dir/new_bin",
                "mode": "100644",
                "type": "blob",
                "sha": "mocked_sha_hash",
            },
        ]
        commit(tmp_path, "main", "dir", commit_message="msg")
        assert len(commit.new_tree_list) == len(expected_tree)
        assert all(item in commit.new_tree_list for item in expected_tree)
        repo.create_tree.assert_called_once()
        assert commit.new_tree_list == repo.create_tree.call_args.args[0]
        assert "base_tree" in repo.create_tree.call_args.kwargs
        repo.create_commit.assert_called_once()

    def test_call__no_changes(self):
        with temporary_dir() as d:
            repo = mock.Mock(spec=Repository)
            repo.tree = mock.Mock(
                return_value=Tree(
                    {"url": "string", "sha": "tree-ish-hash", "tree": []}, None
                )
            )
            commit = CommitDir(repo)
            commit(d, "main", commit_message="msg")
        repo.create_commit.assert_not_called()

    def test_validate_dirs(self):
        repo = mock.Mock(spec=Repository)
        commit = CommitDir(repo)
        with pytest.raises(GithubException):
            commit._validate_dirs("bogus", None)
        _, repo_dir = commit._validate_dirs(".", None)
        assert repo_dir == ""
        _, repo_dir = commit._validate_dirs(".", "./test/")
        assert repo_dir == "test"

    def test_call__error_creating_tree(self):
        with temporary_dir() as d:
            repo = mock.Mock(spec=Repository)
            repo.create_tree.return_value = None
            repo.tree = mock.Mock(
                return_value=Tree(
                    {"url": "string", "sha": "tree-ish-hash", "tree": []}, None
                )
            )
            with open("new", "w") as f:
                f.write("new")
            commit = CommitDir(repo)
            with pytest.raises(GithubException):
                commit(d, "main", commit_message="msg")

    def test_call__error_creating_commit(self):
        with temporary_dir() as d:
            repo = mock.Mock(spec=Repository)
            repo.create_commit.return_value = None
            repo.tree = mock.Mock(
                return_value=Tree(
                    {"url": "string", "sha": "tree-ish-hash", "tree": []}, None
                )
            )
            with open("new", "w") as f:
                f.write("new")
            commit = CommitDir(repo)
            with pytest.raises(GithubException):
                commit(d, "main", commit_message="msg")

    def test_call__error_updating_head(self):
        with temporary_dir() as d:
            repo = mock.Mock(spec=Repository)
            head = mock.Mock()
            head.update.return_value = None
            repo.ref.return_value = head
            repo.tree = mock.Mock(
                return_value=Tree(
                    {"url": "string", "sha": "tree-ish-hash", "tree": []}, None
                )
            )
            with open("new", "w") as f:
                f.write("new")
            commit = CommitDir(repo)
            with pytest.raises(GithubException):
                commit(d, "main", commit_message="msg")

    def test_content_or_sha__dry_run(self, repo):
        commit = CommitDir(repo)
        result = commit._get_content_or_sha(b"content goes here", True)
        assert result == {"content": None}

    def test_content_or_sha__text_content(self, repo):
        commit = CommitDir(repo)
        result = commit._get_content_or_sha(b"content goes here", False)
        assert result == {"content": "content goes here"}

    def test_content_or_sha__binary_sha(self, repo):
        commit = CommitDir(repo)
        repo.create_blob.return_value = "binary-sha"
        result = commit._get_content_or_sha(b"\x9c", False)
        repo.create_blob.assert_called_once_with("nA==", "base64")
        assert result == {"sha": "binary-sha"}

    def test_create_blob__error(self, repo):
        commit = CommitDir(repo)
        repo.create_blob.return_value = None
        with pytest.raises(GithubException):
            commit._get_content_or_sha(b"\x9c", False)



from snowfakery.cci_mapping_files.declaration_parser import SObjectRuleDeclaration

from cumulusci.tasks.bulkdata.extract_dataset_utils.extract_yml import (
    ExtractDeclaration,
)
from cumulusci.tasks.bulkdata.extract_dataset_utils.tests.test_synthesize_extract_declarations import (
    _fake_get_org_schema,
    describe_for,
)
from cumulusci.tasks.bulkdata.generate_mapping_utils.generate_mapping_from_declarations import (
    create_load_mapping_file_from_extract_declarations,
    generate_load_mapping_file,
)
from cumulusci.tasks.bulkdata.mapping_parser import MappingStep
from cumulusci.tasks.bulkdata.step import DataApi, DataOperationType


class TestGenerateLoadMappingFromDeclarations:
    def test_simple_generate_mapping_from_declarations(self, org_config):
        declarations = [
            ExtractDeclaration(sf_object="Account", fields=["Name", "Description"])
        ]
        object_counts = {"Account": 10, "Contact": 0, "Case": 5}
        obj_describes = (
            describe_for("Account"),
            describe_for("Contact"),
            describe_for("Case"),
        )
        with _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            filters=[],
            include_counts=True,
        ) as schema:
            mf = create_load_mapping_file_from_extract_declarations(
                declarations, schema
            )
            assert mf == {
                "Insert Account": {
                    "sf_object": "Account",
                    "table": "Account",
                    "fields": ["Name", "Description"],
                }
            }

    def test_generate_mapping_from_both_kinds_of_declarations(self, org_config):
        declarations = [
            ExtractDeclaration(sf_object="Account", fields=["Name", "Description"]),
            ExtractDeclaration(sf_object="Contact", fields=["FirstName", "LastName"]),
        ]
        object_counts = {"Account": 10, "Contact": 1, "Case": 5}
        obj_describes = (
            describe_for("Account"),
            describe_for("Contact"),
            describe_for("Case"),
        )
        with _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            filters=[],
            include_counts=True,
        ) as schema:
            loading_rules = [
                SObjectRuleDeclaration(sf_object="Account", load_after="Contact")
            ]
            mf = create_load_mapping_file_from_extract_declarations(
                declarations, schema, (), loading_rules
            )
            assert tuple(mf.items()) == tuple(
                {
                    "Insert Contact": {
                        "sf_object": "Contact",
                        "table": "Contact",
                        "fields": ["FirstName", "LastName"],
                    },
                    "Insert Account": {
                        "sf_object": "Account",
                        "table": "Account",
                        "fields": ["Name", "Description"],
                    },
                }.items()
            )

    def test_generate_load_mapping_from_declarations__lookups(self, org_config):
        declarations = [
            ExtractDeclaration(sf_object="Account", fields=["Name", "Description"]),
            ExtractDeclaration(
                sf_object="Contact", fields=["FirstName", "LastName", "AccountId"]
            ),
        ]
        object_counts = {"Account": 10, "Contact": 5, "Case": 5}
        obj_describes = (
            describe_for("Account"),
            describe_for("Contact"),
            describe_for("Case"),
        )
        with _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            filters=[],
            include_counts=True,
        ) as schema:
            mf = create_load_mapping_file_from_extract_declarations(
                declarations, schema
            )
            assert mf == {
                "Insert Account": {
                    "sf_object": "Account",
                    "table": "Account",
                    "fields": ["Name", "Description"],
                },
                "Insert Contact": {
                    "sf_object": "Contact",
                    "table": "Contact",
                    "fields": ["FirstName", "LastName"],
                    "lookups": {
                        "AccountId": {"table": ["Account"], "key_field": "AccountId"}
                    },
                },
            }

    def test_generate_load_mapping_from_declarations__polymorphic_lookups(
        self, org_config
    ):
        """Generate correct mapping file for sobjects with polymorphic lookup fields"""
        declarations = [
            ExtractDeclaration(sf_object="Account", fields=["Name", "Description"]),
            ExtractDeclaration(
                sf_object="Contact", fields=["FirstName", "LastName", "AccountId"]
            ),
            ExtractDeclaration(sf_object="Lead", fields=["LastName", "Company"]),
            ExtractDeclaration(sf_object="Event", fields=["Subject", "WhoId"]),
        ]
        object_counts = {"Account": 10, "Contact": 5, "Case": 5, "Event": 2, "Lead": 1}
        obj_describes = (
            describe_for("Account"),
            describe_for("Contact"),
            describe_for("Case"),
            describe_for("Event"),
            describe_for("Lead"),
        )
        with _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            filters=[],
            include_counts=True,
        ) as schema:
            mf = create_load_mapping_file_from_extract_declarations(
                declarations, schema
            )
            assert mf == {
                "Insert Account": {
                    "sf_object": "Account",
                    "table": "Account",
                    "fields": ["Name", "Description"],
                },
                "Insert Contact": {
                    "sf_object": "Contact",
                    "table": "Contact",
                    "fields": ["FirstName", "LastName"],
                    "lookups": {
                        "AccountId": {"table": ["Account"], "key_field": "AccountId"}
                    },
                },
                "Insert Lead": {
                    "sf_object": "Lead",
                    "table": "Lead",
                    "fields": ["LastName", "Company"],
                },
                "Insert Event": {
                    "sf_object": "Event",
                    "table": "Event",
                    "fields": ["Subject"],
                    "lookups": {
                        "WhoId": {"table": ["Contact", "Lead"], "key_field": "WhoId"}
                    },
                },
            }

    def test_generate_load_mapping_from_declarations__circular_lookups(
        self, org_config
    ):
        declarations = [
            ExtractDeclaration(
                sf_object="Account",
                fields=["Name", "Description", "Primary_Contact__c"],
            ),
            ExtractDeclaration(
                sf_object="Contact", fields=["FirstName", "LastName", "AccountId"]
            ),
        ]
        object_counts = {"Account": 10, "Contact": 5, "Case": 5}
        obj_describes = (
            describe_for("Account"),
            describe_for("Contact"),
            describe_for("Case"),
        )
        with _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            filters=[],
            include_counts=True,
        ) as schema:
            mf = create_load_mapping_file_from_extract_declarations(
                declarations, schema
            )
            assert mf == {
                "Insert Account": {
                    "fields": ["Name", "Description"],
                    "lookups": {
                        "Primary_Contact__c": {
                            "after": "Insert Contact",
                            "key_field": "Primary_Contact__c",
                            "table": ["Contact"],
                        }
                    },
                    "sf_object": "Account",
                    "table": "Account",
                },
                "Insert Contact": {
                    "sf_object": "Contact",
                    "table": "Contact",
                    "fields": ["FirstName", "LastName"],
                    "lookups": {
                        "AccountId": {"table": ["Account"], "key_field": "AccountId"}
                    },
                },
            }, mf

    def test_generate_load_mapping__with_load_declarations(self, org_config):

        mapping_steps = [
            MappingStep(sf_object="Account"),
            MappingStep(sf_object="Contact"),
        ]
        load_declarations = [
            SObjectRuleDeclaration(sf_object="Account", api="rest"),
            SObjectRuleDeclaration(sf_object="Contact", api="bulk"),
        ]
        mf = generate_load_mapping_file(
            mapping_steps,
            intertable_dependencies=set(),
            load_declarations=load_declarations,
        )
        assert mf == {
            "Insert Account": {
                "sf_object": "Account",
                "api": DataApi.REST,
                "table": "Account",
            },
            "Insert Contact": {
                "sf_object": "Contact",
                "api": DataApi.BULK,
                "table": "Contact",
            },
        }, mf

    def test_generate_load_mapping__with_upserts(self, org_config):

        mapping_steps = [
            MappingStep(sf_object="Account"),
            MappingStep(
                sf_object="Account",
                action="UPSERT",
                update_key=["Name"],
                fields=["Name"],
            ),
            MappingStep(
                sf_object="Account",
                action="ETL_UPSERT",
                update_key=["AccountNumber", "Name"],
                fields=["AccountNumber", "Name"],
            ),
            MappingStep(sf_object="Contact"),
        ]

        mf = generate_load_mapping_file(
            mapping_steps,
            intertable_dependencies=set(),
            load_declarations=(),
        )
        assert mf == {
            "Insert Account": {
                "sf_object": "Account",
                "table": "Account",
            },
            "Upsert Account Name": {
                "sf_object": "Account",
                "table": "Account",
                "action": DataOperationType.UPSERT,
                "update_key": ("Name",),
                "fields": ["Name"],
            },
            "Etl_Upsert Account AccountNumber_Name": {
                "sf_object": "Account",
                "table": "Account",
                "action": DataOperationType.ETL_UPSERT,
                "update_key": ("AccountNumber", "Name"),
                "fields": ["AccountNumber", "Name"],
            },
            "Insert Contact": {
                "sf_object": "Contact",
                "table": "Contact",
            },
        }, mf


# TODO: Figure out where clauses



import os
import pathlib
import tempfile
from unittest import mock

import pytest

from cumulusci.core.exceptions import CumulusCIException, TaskOptionsError
from cumulusci.tasks.metadata_etl import MetadataOperation
from cumulusci.tasks.salesforce.update_profile import ProfileGrantAllAccess
from cumulusci.tests.util import DummyOrgConfig, create_project_config
from cumulusci.utils import CUMULUSCI_PATH
from cumulusci.utils.xml import metadata_tree

from .util import create_task

ADMIN_PROFILE_BEFORE = b"""<?xml version='1.0' encoding='utf-8'?>
<Profile xmlns="http://soap.sforce.com/2006/04/metadata">
    <applicationVisibilities>
        <application>npsp__Nonprofit_CRM</application>
        <default>true</default>
        <visible>false</visible>
    </applicationVisibilities>
    <classAccesses>
        <apexClass>TestClass</apexClass>
        <enabled>false</enabled>
    </classAccesses>
    <fieldPermissions>
        <field>Account.TestField__c</field>
        <editable>false</editable>
        <readable>false</readable>
    </fieldPermissions>
    <pageAccesses>
        <apexPage>TestPage</apexPage>
        <enabled>false</enabled>
    </pageAccesses>
    <recordTypeVisibilities>
        <recordType>Account.Business_Account</recordType>
        <default>true</default>
        <personAccountDefault>true</personAccountDefault>
        <visible>true</visible>
    </recordTypeVisibilities>
    <recordTypeVisibilities>
        <recordType>Account.HH_Account</recordType>
        <default>false</default>
        <personAccountDefault>false</personAccountDefault>
        <visible>false</visible>
    </recordTypeVisibilities>
    <tabVisibilities>
        <tab>NPSP_Settings</tab>
        <visibility>Hidden</visibility>
    </tabVisibilities>
</Profile>"""

ADMIN_PROFILE_EXPECTED = """<?xml version="1.0" encoding="UTF-8"?>
<Profile xmlns="http://soap.sforce.com/2006/04/metadata">
    <applicationVisibilities>
        <application>npsp__Nonprofit_CRM</application>
        <default>true</default>
        <visible>true</visible>
    </applicationVisibilities>
    <classAccesses>
        <apexClass>TestClass</apexClass>
        <enabled>true</enabled>
    </classAccesses>
    <fieldPermissions>
        <field>Account.TestField__c</field>
        <editable>true</editable>
        <readable>true</readable>
    </fieldPermissions>
    <pageAccesses>
        <apexPage>TestPage</apexPage>
        <enabled>true</enabled>
    </pageAccesses>
    <recordTypeVisibilities>
        <recordType>Account.Business_Account</recordType>
        <default>false</default>
        <personAccountDefault>false</personAccountDefault>
        <visible>true</visible>
    </recordTypeVisibilities>
    <recordTypeVisibilities>
        <recordType>Account.HH_Account</recordType>
        <default>true</default>
        <personAccountDefault>true</personAccountDefault>
        <visible>true</visible>
    </recordTypeVisibilities>
    <tabVisibilities>
        <tab>NPSP_Settings</tab>
        <visibility>DefaultOn</visibility>
    </tabVisibilities>
</Profile>
"""

PACKAGE_XML_BEFORE = """<Package xmlns="http://soap.sforce.com/2006/04/metadata">
    <types>
        <members>*</members>
        <members>Account</members>
        <members>Campaign</members>
        <members>CampaignMember</members>
        <members>Contact</members>
        <members>Lead</members>
        <members>Opportunity</members>
        <name>CustomObject</name>
    </types>
    <types>
        <members>*</members>
        <name>CustomTab</name>
    </types>
    <types>
        <name>Profile</name>
    </types>
    <version>39.0</version>
</Package>"""

PACKAGE_XML_BEFORE__PROFILES = """<Package xmlns="http://soap.sforce.com/2006/04/metadata">
    <types>
        <members>*</members>
        <members>Account</members>
        <name>CustomObject</name>
    </types>
    <types>
        <name>Profile</name>
        <members>Admin</members>
        <members>Test</members>
    </types>
    <version>39.0</version>
</Package>"""

PACKAGE_XML_BEFORE__NO_PROFILES = """<Package xmlns="http://soap.sforce.com/2006/04/metadata">
    <types>
        <members>*</members>
        <members>Account</members>
        <name>CustomObject</name>
    </types>
    <version>39.0</version>
</Package>"""

ADMIN_PROFILE_BEFORE__MULTI_OBJECT_RT = b"""<?xml version='1.0' encoding='utf-8'?>
<Profile xmlns="http://soap.sforce.com/2006/04/metadata">
    <recordTypeVisibilities>
        <recordType>Account.Business_Account</recordType>
        <default>false</default>
        <personAccountDefault>true</personAccountDefault>
        <visible>true</visible>
    </recordTypeVisibilities>
    <recordTypeVisibilities>
        <recordType>Opportunity.Donation</recordType>
        <default>true</default>
        <personAccountDefault>false</personAccountDefault>
        <visible>false</visible>
    </recordTypeVisibilities>
</Profile>"""

ADMIN_PROFILE_EXPECTED__MULTI_OBJECT_RT = """<?xml version="1.0" encoding="UTF-8"?>
<Profile xmlns="http://soap.sforce.com/2006/04/metadata">
    <recordTypeVisibilities>
        <recordType>Account.Business_Account</recordType>
        <default>true</default>
        <personAccountDefault>true</personAccountDefault>
        <visible>true</visible>
    </recordTypeVisibilities>
    <recordTypeVisibilities>
        <recordType>Opportunity.Donation</recordType>
        <default>true</default>
        <personAccountDefault>false</personAccountDefault>
        <visible>false</visible>
    </recordTypeVisibilities>
</Profile>
"""


def test_run_task():
    with tempfile.TemporaryDirectory() as tempdir:
        with open(
            os.path.join(CUMULUSCI_PATH, "cumulusci", "files", "admin_profile.xml"), "w"
        ) as f:
            f.write(PACKAGE_XML_BEFORE)

        retrieve_dir = pathlib.Path(tempdir, "retrieve")
        retrieve_dir.mkdir()
        target_dir = retrieve_dir / "profiles"
        target_dir.mkdir()

        with open(target_dir / "Admin.profile", "wb") as f:
            f.write(ADMIN_PROFILE_BEFORE)

        task = create_task(
            ProfileGrantAllAccess,
            {
                "record_types": [
                    {
                        "record_type": "Account.HH_Account",
                        "default": True,
                        "person_account_default": True,
                    }
                ],
                "include_packaged_objects": False,
            },
        )

        task._retrieve = mock.Mock()
        task._deploy = mock.Mock()
        task._create_directories = mock.Mock()
        task.retrieve_dir = retrieve_dir
        task.deploy_dir = pathlib.Path(tempdir, "deploy")
        task.deploy_dir.mkdir()

        task()

        dest_path = task.deploy_dir / "profiles" / "Admin.profile"
        assert dest_path.exists()

        assert dest_path.read_text() == ADMIN_PROFILE_EXPECTED


def test_transforms_profile():
    task = create_task(
        ProfileGrantAllAccess,
        {
            "record_types": [
                {
                    "record_type": "Account.HH_Account",
                    "default": True,
                    "person_account_default": True,
                }
            ],
        },
    )

    inbound = metadata_tree.fromstring(ADMIN_PROFILE_BEFORE)
    outbound = task._transform_entity(inbound, "Admin")

    xml_output = outbound.tostring(xml_declaration=True)

    assert xml_output == ADMIN_PROFILE_EXPECTED


def test_transforms_profile__multi_object_rt():
    task = create_task(
        ProfileGrantAllAccess,
        {
            "record_types": [
                {
                    "record_type": "Account.Business_Account",
                    "default": True,
                    "person_account_default": True,
                }
            ],
        },
    )

    inbound = metadata_tree.fromstring(ADMIN_PROFILE_BEFORE__MULTI_OBJECT_RT)
    outbound = task._transform_entity(inbound, "Admin")

    xml_output = outbound.tostring(xml_declaration=True)

    assert xml_output == ADMIN_PROFILE_EXPECTED__MULTI_OBJECT_RT


def test_throws_exception_record_type_not_found():
    task = create_task(
        ProfileGrantAllAccess,
        {
            "record_types": [{"record_type": "DOESNT_EXIST"}],
        },
    )

    with pytest.raises(TaskOptionsError):
        task._transform_entity(metadata_tree.fromstring(ADMIN_PROFILE_BEFORE), "Admin")


def test_transforms_profile_with_layouts():
    task = create_task(
        ProfileGrantAllAccess,
        {
            "record_types": [
                {
                    "record_type": "Account.HH_Account",
                    "page_layout": "Account-Account Layout",
                },
                {
                    "record_type": "Account.Business_Account",
                    "page_layout": "Account-Account Layout",
                    "default": True,
                },
            ],
        },
    )

    # To support coverage for when pulling metadata with layout definitions
    profile_before = ADMIN_PROFILE_BEFORE.replace(
        b"</Profile>",
        b"""\n      <layoutAssignments>
            <recordType>Account.Business_Account</recordType>
            <layout>Account-Business Account Layout</layout>
        </layoutAssignments>
    </Profile>""",
    )

    inbound = metadata_tree.fromstring(profile_before)
    assert len(inbound.findall("layoutAssignments")) == 1

    # Verify the expected input includes the layoutAssignment element and
    # the original layout
    assert (
        inbound.find("layoutAssignments", recordType="Account.Business_Account")
        .find("layout")
        .text
        == "Account-Business Account Layout"
    )

    # Run the transformation and assert that there are two layoutAssignments
    outbound = task._transform_entity(inbound, "Admin")
    assert len(outbound.findall("layoutAssignments")) == 2

    # Verify that the original layout assignment was updated
    assignments = outbound.find(
        "layoutAssignments", recordType="Account.Business_Account"
    )
    assert assignments is not None
    assert assignments.find("layout").text == "Account-Account Layout"

    assignments = outbound.find("layoutAssignments", recordType="Account.HH_Account")
    assert assignments is not None
    assert assignments.find("layout").text == "Account-Account Layout"

    # assert that the nodes are included in the xml output as expected
    xml_output = outbound.tostring(xml_declaration=True)
    assert xml_output.count("<layoutAssignments>") == 2


def test_expand_package_xml():
    task = create_task(ProfileGrantAllAccess, {"api_names": ["Admin"]})
    task.tooling = mock.Mock()
    task.tooling.query_all.return_value = {
        "totalSize": 2,
        "records": [
            {"DeveloperName": "Test", "NamespacePrefix": "ns"},
            {"DeveloperName": "Foo_Bar", "NamespacePrefix": "fb"},
        ],
    }

    package_xml = metadata_tree.fromstring(PACKAGE_XML_BEFORE)
    task._expand_package_xml(package_xml)
    types = package_xml.find("types", name="CustomObject")
    assert "ns__Test__c" in {elem.text for elem in types.findall("members")}
    assert "fb__Foo_Bar__c" in {elem.text for elem in types.findall("members")}


def test_expand_package_xml_objects():
    task = create_task(
        ProfileGrantAllAccess,
        {"api_names": ["Admin"], "record_types": [{"record_type": "Case.Case"}]},
    )
    task.tooling = mock.Mock()
    package_xml = metadata_tree.fromstring(PACKAGE_XML_BEFORE)
    task._expand_package_xml_objects(package_xml)
    types = package_xml.find("types", name="CustomObject")
    assert "Case" in {elem.text for elem in types.findall("members")}


def test_expand_package_xml_objects_no_record_types():
    task = create_task(
        ProfileGrantAllAccess,
        {"api_names": ["Admin"]},
    )
    task.tooling = mock.Mock()
    package_xml = metadata_tree.fromstring(PACKAGE_XML_BEFORE)
    task._expand_package_xml_objects(package_xml)
    types = package_xml.find("types", name="CustomObject")
    assert "Case" not in {elem.text for elem in types.findall("members")}


def test_expand_package_xml__broken_package():
    task = create_task(ProfileGrantAllAccess, {"api_names": ["Admin"]})
    task.tooling = mock.Mock()
    task.tooling.query_all.return_value = {
        "totalSize": 2,
        "records": [
            {"DeveloperName": "Test", "NamespacePrefix": "ns"},
            {"DeveloperName": "Foo_Bar", "NamespacePrefix": "fb"},
        ],
    }

    package_xml = metadata_tree.fromstring(PACKAGE_XML_BEFORE)
    types = package_xml.find("types", name="CustomObject")
    types.find("name").text = "NotCustomObject"
    with pytest.raises(CumulusCIException):
        task._expand_package_xml(package_xml)


def test_expand_profile_members():
    task = create_task(
        ProfileGrantAllAccess,
        {
            "api_names": ["Admin", "%%%NAMESPACE%%%Continuous Integration"],
            "namespace_inject": "ns",
            "managed": True,
        },
    )
    package_xml = metadata_tree.fromstring(PACKAGE_XML_BEFORE)

    task._expand_profile_members(package_xml)

    types = package_xml.find("types", name="Profile")
    assert {elem.text for elem in types.findall("members")} == {
        "Admin",
        "ns__Continuous Integration",
    }


def test_expand_profile_members__no_api_names():
    task = create_task(
        ProfileGrantAllAccess,
        {
            "namespace_inject": "ns",
            "managed": True,
        },
    )
    package_xml = metadata_tree.fromstring(PACKAGE_XML_BEFORE)

    task._expand_profile_members(package_xml)

    types = package_xml.find("types", name="Profile")
    assert {elem.text for elem in types.findall("members")} == {"Admin"}


def test_expand_profile_members__existing_entries():
    task = create_task(
        ProfileGrantAllAccess,
        {
            "api_names": ["Admin", "%%%NAMESPACE%%%Continuous Integration"],
            "namespace_inject": "ns",
            "managed": True,
        },
    )
    package_xml = metadata_tree.fromstring(PACKAGE_XML_BEFORE__PROFILES)

    task._expand_profile_members(package_xml)

    types = package_xml.find("types", name="Profile")

    assert {elem.text for elem in types.findall("members")} == {
        "Admin",
        "Test",
        "ns__Continuous Integration",
    }

    assert task.api_names == {"Admin", "ns__Continuous Integration", "Test"}


def test_expand_profile_members__no_profile_section():
    task = create_task(
        ProfileGrantAllAccess,
        {
            "api_names": ["Admin", "%%%NAMESPACE%%%Continuous Integration"],
            "namespace_inject": "ns",
            "managed": True,
        },
    )
    package_xml = metadata_tree.fromstring(PACKAGE_XML_BEFORE__NO_PROFILES)

    task._expand_profile_members(package_xml)

    types = package_xml.find("types", name="Profile")

    assert {elem.text for elem in types.findall("members")} == {
        "Admin",
        "ns__Continuous Integration",
    }

    assert task.api_names == {"Admin", "ns__Continuous Integration"}


def test_expand_profile_members__namespaced_org():
    project_config = create_project_config(namespace="ns")
    task = create_task(
        ProfileGrantAllAccess,
        {
            "api_names": ["Admin", "%%%NAMESPACED_ORG%%%Continuous Integration"],
            "namespace_inject": "ns",
            "namespaced_org": True,
            "managed": False,
        },
        project_config,
    )
    package_xml = metadata_tree.fromstring(PACKAGE_XML_BEFORE)

    task._expand_profile_members(package_xml)

    types = package_xml.find("types", name="Profile")
    assert {elem.text for elem in types.findall("members")} == {
        "Admin",
        "ns__Continuous Integration",
    }


def test_init_options__general():
    project_config = create_project_config(namespace="ns")
    task = create_task(
        ProfileGrantAllAccess, {"managed": "true"}, project_config=project_config
    )
    assert task.options["managed"]
    assert not task.options["namespaced_org"]
    assert task.options["namespace_inject"] == "ns"
    assert task.namespace_prefixes == {"managed": "ns__", "namespaced_org": ""}

    task = create_task(
        ProfileGrantAllAccess,
        {"namespaced_org": "true", "managed": False},
        project_config=project_config,
    )
    assert not task.options["managed"]
    assert task.options["namespaced_org"]
    assert task.options["namespace_inject"] == "ns"
    assert task.namespace_prefixes == {"managed": "", "namespaced_org": "ns__"}

    task = create_task(ProfileGrantAllAccess, {})
    assert not task.options["managed"]
    assert not task.options["namespaced_org"]
    assert task.options["namespace_inject"] is None
    assert task.namespace_prefixes == {"managed": "", "namespaced_org": ""}


def test_init_options__api_names():
    task = create_task(
        ProfileGrantAllAccess,
        {
            "api_names": ["Admin", "%%%NAMESPACE%%%Continuous Integration"],
            "namespace_inject": "ns",
            "managed": True,
        },
    )
    assert task.api_names == {"Admin", "ns__Continuous Integration"}

    task = create_task(
        ProfileGrantAllAccess,
        {
            "api_names": ["Admin", "%%%NAMESPACE%%%Continuous Integration"],
            "namespace_inject": "ns",
            "managed": False,
        },
    )
    assert task.api_names == {"Admin", "Continuous Integration"}

    task = create_task(
        ProfileGrantAllAccess, {"profile_name": "Continuous Integration"}
    )
    assert task.api_names == {"Continuous Integration"}

    task = create_task(ProfileGrantAllAccess, {})
    assert task.api_names == {"Admin"}

    task = create_task(ProfileGrantAllAccess, {"package_xml": "lib/admin_profile.xml"})
    assert task.api_names == set()


def test_init_options__include_packaged_objects():
    pc = create_project_config()
    pc.minimum_cumulusci_version = "4.0.0"
    task = create_task(
        ProfileGrantAllAccess,
        {"package_xml": "lib/admin_profile.xml"},
        project_config=pc,
    )
    assert task.package_xml_path == "lib/admin_profile.xml"
    assert not task.options["include_packaged_objects"]

    task = create_task(ProfileGrantAllAccess, {}, project_config=pc)
    assert task.options["include_packaged_objects"]
    assert task.package_xml_path == os.path.join(
        CUMULUSCI_PATH, "cumulusci", "files", "admin_profile.xml"
    )

    pc.minimum_cumulusci_version = "2.0.0"
    task = create_task(ProfileGrantAllAccess, {}, project_config=pc)
    assert not task.options["include_packaged_objects"]

    task = create_task(
        ProfileGrantAllAccess, {"include_packaged_objects": True}, project_config=pc
    )
    assert task.options["include_packaged_objects"]


def test_init_options__namespace_injection():
    pc = create_project_config(namespace="ns")
    org_config = DummyOrgConfig({"namespace": "ns"})
    org_config._installed_packages = {"ns": None}
    task = create_task(
        ProfileGrantAllAccess, {}, project_config=pc, org_config=org_config
    )
    assert task.options["namespace_inject"] == "ns"
    assert task.options["namespaced_org"]
    assert task.options["managed"]


def test_generate_package_xml__retrieve():
    with tempfile.TemporaryDirectory() as tmp_dir:
        admin_profile = os.path.join(tmp_dir, "admin_profile.xml")
        with open(admin_profile, "w") as f:
            f.write(PACKAGE_XML_BEFORE)
        with open(
            os.path.join(CUMULUSCI_PATH, "cumulusci", "files", "admin_profile.xml"), "w"
        ) as f:
            f.write(PACKAGE_XML_BEFORE)

        task = create_task(
            ProfileGrantAllAccess,
            {"package_xml": admin_profile, "include_packaged_objects": False},
        )

        task._expand_profile_members = mock.Mock()
        task._expand_package_xml = mock.Mock()
        task._generate_package_xml(MetadataOperation.RETRIEVE)
        task._expand_package_xml.assert_not_called()

        task = create_task(
            ProfileGrantAllAccess,
            {"package_xml": admin_profile, "include_packaged_objects": True},
        )

        task._expand_profile_members = mock.Mock()
        task._expand_package_xml = mock.Mock()
        task._generate_package_xml(MetadataOperation.RETRIEVE)
        task._expand_package_xml.assert_called_once()

        task = create_task(ProfileGrantAllAccess, {"include_packaged_objects": True})

        task._expand_profile_members = mock.Mock()
        task._expand_package_xml = mock.Mock()
        task._generate_package_xml(MetadataOperation.RETRIEVE)
        task._expand_package_xml.assert_called_once()

        task = create_task(ProfileGrantAllAccess, {"include_packaged_objects": False})

        task._expand_profile_members = mock.Mock()
        task._expand_package_xml = mock.Mock()
        task._generate_package_xml(MetadataOperation.RETRIEVE)
        task._expand_package_xml.assert_not_called()



import pytest

from cumulusci.core.config import ServiceConfig
from cumulusci.core.config.marketing_cloud_service_config import (
    MarketingCloudServiceConfig,
)
from cumulusci.tests.util import create_project_config

from ..mc_constants import MC_API_VERSION


@pytest.fixture
def mc_project_config():
    project_config = create_project_config()
    project_config.keychain.set_service(
        "oauth2_client",
        "test",
        ServiceConfig(
            {
                "client_id": "MC_CLIENT_ID",
                "client_secret": "BOGUS",
                "auth_uri": f"https://TSSD.auth.marketingcloudapis.com/{MC_API_VERSION}/authorize",
                "token_uri": f"https://TSSD.auth.marketingcloudapis.com/{MC_API_VERSION}/token",
                "callback_url": "https://127.0.0.1:8080/",
            },
            "test",
            project_config.keychain,
        ),
        False,
    )
    project_config.keychain.set_service(
        "marketing_cloud",
        "test",
        MarketingCloudServiceConfig(
            {
                "oauth2_client": "test",
                "refresh_token": "REFRESH",
                "rest_instance_url": "https://TSSD.auth.marketingcloudapis.com/",
                "soap_instance_url": "https://TSSD.soap.marketingcloudapis.com/",
            },
            "test",
            project_config.keychain,
        ),
        False,
    )
    return project_config



from contextlib import contextmanager
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import mock

import pytest
import responses

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.tasks.bulkdata.extract_dataset_utils.tests.test_synthesize_extract_declarations import (
    _fake_get_org_schema,
    describe_for,
)
from cumulusci.tasks.sample_data.capture_sample_data import CaptureSampleData
from cumulusci.tests.util import CURRENT_SF_API_VERSION


@contextmanager
def setup_test(org_config):
    object_counts = {"Account": 6, "Contact": 1, "Opportunity": 5}
    obj_describes = (
        describe_for("Account"),
        describe_for("Contact"),
        describe_for("Opportunity"),
    )
    with mock.patch.object(
        type(org_config), "is_person_accounts_enabled", False
    ), mock.patch(
        "cumulusci.tasks.sample_data.capture_sample_data.get_org_schema",
        lambda _sf, org_config, **kwargs: _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            included_objects=["Account", "Contact", "Opportunity"],
            **kwargs,
        ),
    ), responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            f"{org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/tooling/sobjects",
            json={"sobjects": [{"name": "WebLink"}]},
            status=200,
        )
        yield


class TestCaptureDatasets:
    @mock.patch("cumulusci.tasks.sample_data.capture_sample_data.Dataset")
    def test_simple_extract(
        self,
        Dataset,
        create_task,
        org_config,
    ):
        with setup_test(org_config):
            task = create_task(CaptureSampleData, {})
            task()
            print("Dataset.mock_calls", Dataset.mock_calls)
            # default dataset should created
            Dataset.assert_any_call("default", mock.ANY, mock.ANY, org_config, mock.ANY)
            # and extracted
            Dataset().__enter__().extract.assert_called_with(
                {}, task.logger, None, mock.ANY, None
            )

    @mock.patch("cumulusci.tasks.sample_data.capture_sample_data.Dataset")
    def test_named_extract(
        self,
        Dataset,
        create_task,
        org_config,
    ):
        with setup_test(org_config), TemporaryDirectory() as t:
            extraction_definition = Path(t) / "test_extract_definition"
            with open(extraction_definition, "w") as f:
                f.write("")  # Doesn't matter. We won't parse it

            loading_rules = Path(t) / "test_load_definition"
            with open(loading_rules, "w") as f:
                f.write("[]")  # Doesn't matter. We won't parse it

            Dataset().__enter__().path.exists.return_value = False
            task = create_task(
                CaptureSampleData,
                {
                    "dataset": "mydataset",
                    "extraction_definition": extraction_definition,
                    "loading_rules": loading_rules,
                },
            )
            task()
            # named dataset should create
            Dataset.assert_any_call(
                "mydataset", mock.ANY, mock.ANY, org_config, mock.ANY
            )
            Dataset().__enter__().create.assert_called_with()
            # and extracted
            Dataset().__enter__().extract.assert_called_with(
                {}, task.logger, extraction_definition, mock.ANY, loading_rules
            )

    @mock.patch("cumulusci.tasks.sample_data.capture_sample_data.Dataset")
    def test_option_error(
        self,
        Dataset,
        create_task,
        org_config,
    ):
        task = create_task(
            CaptureSampleData,
            {
                "dataset": "mydataset",
                "extraction_definition": Path("xyzzy.bbb.yml"),
            },
        )
        with pytest.raises(TaskOptionsError, match="xyzzy.bbb.yml"):
            task()



from datetime import datetime

from cumulusci.core.github import get_tag_by_name
from cumulusci.tasks.github.base import BaseGithubTask


class CloneTag(BaseGithubTask):
    task_options = {  # TODO: should use `class Options instead`
        "src_tag": {
            "description": "The source tag to clone.  Ex: beta/1.0-Beta_2",
            "required": True,
        },
        "tag": {
            "description": "The new tag to create by cloning the src tag.  Ex: release/1.0",
            "required": True,
        },
    }

    def _run_task(self):
        src_tag_name = self.options["src_tag"]
        repo = self.get_repo()
        src_tag = get_tag_by_name(repo, src_tag_name)

        tag = repo.create_tag(
            tag=self.options["tag"],
            message=f"Cloned from {src_tag_name}",
            sha=src_tag.sha,
            obj_type="commit",
            tagger={
                "name": self.github_config.username,
                "email": self.github_config.email,
                "date": f"{datetime.utcnow().isoformat()}Z",
            },
        )
        self.logger.info(f"Tag {self.options['tag']} created by cloning {src_tag_name}")

        return tag



import json
import uuid
import zipfile
from collections import defaultdict
from enum import Enum
from pathlib import Path
from typing import Dict, Optional

import requests

from cumulusci.core.debug import get_debug_mode
from cumulusci.core.exceptions import DeploymentException
from cumulusci.core.utils import process_list_of_pairs_dict_arg
from cumulusci.utils import temporary_dir
from cumulusci.utils.http.requests_utils import safe_json_from_response

from .base import BaseMarketingCloudTask

MCPM_BASE_ENDPOINT = "https://spf.{}.marketingcloudapps.com/api"

MCPM_JOB_ID_HEADER = "x-mcpm-job-id"

PAYLOAD_CONFIG_VALUES = {"preserveCategories": True}

PAYLOAD_NAMESPACE_VALUES = {
    "category": "",
    "prepend": "",
    "append": "",
    "timestamp": True,
}

IN_PROGRESS_STATUSES = ("NOT_STARTED", "IN_PROGRESS")
FINISHED_STATUSES = ("DONE",)
ERROR_STATUSES = ("FATAL_ERROR", "ERROR")


UNKNOWN_STATUS_MESSAGE = "Received unknown deploy status: {}"


class PollAction(Enum):
    validating = "VALIDATING"
    deploying = "DEPLOYING"


class MarketingCloudDeployTask(BaseMarketingCloudTask):
    # This task executes multiple polling loops.
    # This enables the task to determine which endpoints should be polled.
    current_action: Optional[PollAction] = None
    validation_not_found_count = 0
    job_name: Optional[str]

    task_options = {
        "package_zip_file": {
            "description": "Path to the package zipfile that will be deployed.",
            "required": True,
        },
        "custom_inputs": {
            "description": "Specify custom inputs to the deployment task. Takes a mapping from input key to input value (e.g. 'companyName:Acme,companyWebsite:https://www.salesforce.org:8080').",
            "required": False,
        },
        "name": {
            "description": "The name to give to this particular deploy call. Defaults to a universally unique identifier.",
            "required": False,
        },
        "endpoint": {
            "description": "Override the default endpoint for the Marketing Cloud package manager API (optional)",
            "required": False,
        },
    }

    def _init_options(self, kwargs):
        self.debug_mode = get_debug_mode()
        super()._init_options(kwargs)
        custom_inputs = self.options.get("custom_inputs")
        self.custom_inputs = (
            process_list_of_pairs_dict_arg(custom_inputs) if custom_inputs else None
        )

    def _run_task(self):
        # Marketing Cloud validation _requires_ a 2-second static polling interval
        self.poll_interval_s = 2

        pkg_zip_file = Path(self.options["package_zip_file"])
        if not pkg_zip_file.is_file():
            self.logger.error(f"Package zip file not valid: {pkg_zip_file.name}")
            return

        with temporary_dir(chdir=False) as temp_dir:
            with zipfile.ZipFile(pkg_zip_file) as zf:
                zf.extractall(temp_dir)
                payload = self._construct_payload(Path(temp_dir), self.custom_inputs)

        # These initialization steps are not done in _init_options()
        # because they use MC authorization. We don't want to freeze
        # the responses.
        endpoint_option = self.options.get("endpoint")
        self.endpoint = endpoint_option or MCPM_BASE_ENDPOINT.format(
            self.get_mc_stack_key()
        )

        self.headers = {
            "Authorization": f"Bearer {self.mc_config.access_token}",
            "SFMC-TSSD": self.mc_config.tssd,
        }

        self._validate_package(payload)
        self._reset_poll()

        payload = self._update_payload_entities_with_actions(
            payload,
        )
        self._deploy_package(payload)

    def _construct_payload(self, dir_path, custom_inputs=None):
        dir_path = Path(dir_path)
        assert dir_path.is_dir(), "package_directory must be a directory"

        payload = defaultdict(lambda: defaultdict(dict))
        payload["namespace"] = PAYLOAD_NAMESPACE_VALUES
        payload["config"] = PAYLOAD_CONFIG_VALUES
        payload["name"] = self.job_name = self.options.get("name", str(uuid.uuid4()))  # type: ignore

        try:
            with open(dir_path / "info.json", "r") as f:
                info_json = json.load(f)
                model_version = info_json.get("modelVersion", "1")

                self.logger.debug(
                    f"Setting Marketing Cloud Package Manager modelVersion to: {model_version}"
                )
                payload["modelVersion"] = model_version

            with open(dir_path / "references.json", "r") as f:
                payload["references"] = json.load(f)

            with open(dir_path / "input.json", "r") as f:
                payload["input"] = json.load(f)

            entities_dir = Path(f"{dir_path}/entities")
            for item in entities_dir.glob("**/*.json"):
                if item.is_file():
                    entity_name = item.parent.name
                    entity_id = item.stem
                    with open(item, "r") as f:
                        payload["entities"][entity_name][entity_id] = json.load(f)
        except FileNotFoundError as e:
            msg = f"Expected file not found in provided package zip file: {e.filename.split('/')[-1]}"
            raise DeploymentException(msg)

        if custom_inputs:
            payload = self._add_custom_inputs_to_payload(custom_inputs, payload)

        if self.debug_mode:  # pragma: nocover
            self.logger.debug(f"Payload:\n{json.dumps(payload)}")

        return payload

    def _add_custom_inputs_to_payload(self, custom_inputs, payload):
        for input_name, value in custom_inputs.items():
            found = False
            for input in payload["input"]:
                if input["key"] == input_name:
                    input["value"] = value
                    found = True
                    break

            if not found:
                raise DeploymentException(
                    f"Custom input of key {input_name} not found in package."
                )

        return payload

    def _validate_package(self, payload: Dict) -> None:
        """Sends the payload to MC for validation.
        Returns a dict of allowable actions for the target MC instance."""
        self.current_action = PollAction.validating
        self.logger.info(f"Validating package at: {self.endpoint}/validate")
        assert self.job_name
        response = requests.post(
            f"{self.endpoint}/validate",
            json=payload,
            headers={
                MCPM_JOB_ID_HEADER: self.job_name,
                **self.headers,
            },
        )
        response.raise_for_status()
        response_data = safe_json_from_response(response)
        self.validate_id = response_data["id"]
        self.logger.info(f"Started package validation with Id: {self.validate_id}")
        self._poll()

    def _update_payload_entities_with_actions(self, payload: Dict) -> Dict:
        """Include available entity action returned from the validation
        endpoint into the payload used for package deployment."""

        for entity_type in payload["entities"]:
            this_entity_type = payload["entities"][entity_type]
            for entity_id in this_entity_type:
                this_entity = payload["entities"][entity_type][entity_id]
                action = self.action_for_entity(entity_type, entity_id)
                this_entity["action"] = action

        if self.debug_mode:
            self.logger.debug(f"Payload updated with actions:\n{json.dumps(payload)}")

        return payload

    def action_for_entity(self, entity: str, entity_id: str) -> Optional[Dict]:
        """Fetch the corresponding action for the given entity with the specified Id"""
        try:
            for action_info in self.validation_response["entities"][entity][entity_id][
                "actions"
            ].values():
                if action_info["available"]:
                    return action_info
        except KeyError:
            # if no actions are defined for this entity just move on
            pass

    def _deploy_package(self, payload: Dict) -> None:
        self.current_action = PollAction.deploying
        self.logger.info(f"Deploying package to: {self.endpoint}/deployments")
        response = requests.post(
            f"{self.endpoint}/deployments",
            json=payload,
            headers=self.headers,
        )
        response.raise_for_status()
        response_data = safe_json_from_response(response)
        self.deploy_id = response_data["id"]
        self.logger.info(f"Started package deploy with Id: {self.validate_id}")
        self._poll()

    def _poll_action(self) -> None:
        """
        Poll based on the current action being taken.
        """
        if self.current_action == PollAction.validating:
            self._poll_validating()
        elif self.current_action == PollAction.deploying:
            self._poll_deploying()
        else:
            raise Exception(
                f"PollAction {self.current_action} does not have a polling handler defined."
            )

    def _poll_validating(self) -> None:
        assert self.job_name
        response = requests.get(
            f"{self.endpoint}/validate/{self.validate_id}",
            headers={MCPM_JOB_ID_HEADER: self.job_name, **self.headers},
        )
        response_data = safe_json_from_response(response)
        validation_status = response_data["status"]
        self.logger.info(f"Validation status is: {validation_status}")

        # Handle eccentricities of Marketing Cloud validation polling.
        # We may get back a NOT_FOUND result if our request is routed
        # to the wrong pod by MC. This should be fixed, but hasn't been
        # deployed to all MC stacks yet.

        # Observed behavior is that response switches between NOT_FOUND
        # and IN_PROGRESS. We'll track these responses and fail the job
        # only if we consistently receive NOT_FOUND.

        # TODO: add a timeout.
        if validation_status == "NOT_FOUND":
            self.validation_not_found_count += 1
            if self.validation_not_found_count > 10:
                raise DeploymentException(
                    f"Unable to find status on validation: {self.validate_id}"
                )
        else:
            # Reset if we get back a result other than NOT_FOUND
            self.validation_not_found_count = 0
            if validation_status not in IN_PROGRESS_STATUSES:
                self.poll_complete = True
                if self.debug_mode:  # pragma: nocover
                    self.logger.debug(
                        f"Validation Response:\n{json.dumps(response_data)}"
                    )
                self.validation_response = response_data

    def _poll_deploying(self) -> None:
        response = requests.get(
            f"{self.endpoint}/deployments/{self.deploy_id}", headers=self.headers
        )
        response_data = safe_json_from_response(response)
        deploy_status = response_data["status"]
        self.logger.info(f"Deployment status is: {deploy_status}")

        if deploy_status not in IN_PROGRESS_STATUSES:
            self._process_completed_deploy(response_data)

    def _poll_update_interval(self):
        # Marketing Cloud validation _requires_ a 2-second static polling interval.
        # Override the base class to remove backoff.
        pass

    def _process_completed_deploy(self, response_data: Dict):
        deploy_status = response_data["status"]
        assert (
            deploy_status != IN_PROGRESS_STATUSES
        ), "Deploy should be in a completed state before processing."

        self.poll_complete = True
        if deploy_status in FINISHED_STATUSES:
            self._validate_response(response_data)
        elif deploy_status in ERROR_STATUSES:
            self._report_error(response_data)
        else:
            self.logger.error(UNKNOWN_STATUS_MESSAGE.format(deploy_status))
            self._report_error(response_data)

    def _validate_response(self, deploy_info: Dict) -> None:
        """Checks for any errors present in the response to the deploy request.
        Displays errors if present, else informs use that the deployment was successful.
        """
        has_error = False
        for entity, info in deploy_info["entities"].items():
            if not info:
                continue

            for entity_id, info in info.items():
                if info["status"] not in ("SUCCESS", "REUSED"):
                    has_error = True
                    self.logger.error(
                        f"Failed to deploy {entity}/{entity_id}. Status: {info['status']}. Issues: {info['issues']}"
                    )

        if has_error:
            raise DeploymentException("Marketing Cloud reported deployment failures.")

        self.logger.info(f"Deployment ({self.deploy_id}) completed successfully.")

    def _report_error(self, response_data: Dict) -> None:
        deploy_status = response_data["status"]
        self.logger.error(
            f"Received status of: {deploy_status}. Received the following data from Marketing Cloud:\n{json.dumps(response_data)}"
        )
        raise DeploymentException(
            f"Marketing Cloud deploy finished with status of: {deploy_status}"
        )



import itertools
import typing as T
from itertools import filterfalse, tee


def iterate_in_chunks(n: int, iterable: T.Iterable) -> T.Iterable[T.Tuple]:
    """Split an iterable into chunks of size 'n'."""
    it = iter(iterable)
    while True:
        chunk = tuple(itertools.islice(it, n))
        if not chunk:
            return
        yield chunk


# From https://docs.python.org/3/library/itertools.html
def partition(pred, iterable):
    "Use a predicate to partition entries into false entries and true entries"
    # partition(is_odd, range(10)) --> 0 2 4 6 8   and  1 3 5 7 9
    t1, t2 = tee(iterable)
    return filterfalse(pred, t1), filter(pred, t2)



import typing as T

from cumulusci.core import exceptions as exc


class RunUntilBase:
    # subclasses need to fill in these two fields
    sobject_name: str = None
    target: int = None
    gap: int = None

    def set_target_and_gap(
        self, sobject_name: T.Optional[str], num_as_str: T.Union[str, int]
    ):
        self.sobject_name = sobject_name
        try:
            self.target = self.gap = int(num_as_str)
        except (TypeError, ValueError):
            raise exc.TaskOptionsError(f"{num_as_str} is not a number")

    def split_pair(self, param):
        parts = param.split(":")
        if len(parts) != 2:
            raise exc.TaskOptionsError(
                f"{param} is in the wrong format for {self.option_name}"
            )
        return parts

    @property
    def nothing_to_do(self):
        return self.gap <= 0


class RunUntilRecipeRepeated(RunUntilBase):
    option_name = "--run-until-recipe-repeated"
    nothing_to_do_because = "You asked for zero repetitions."

    def __init__(self, sf, param):
        self.set_target_and_gap(None, param)


class RunUntilRecordsLoaded(RunUntilBase):
    option_name = "--run-until-records-loaded"
    nothing_to_do_because = "You asked for zero records."

    def __init__(self, sf, param):
        parts = self.split_pair(param)
        self.set_target_and_gap(*parts)


class RunUntilRecordInOrg(RunUntilRecordsLoaded):
    option_name = "--run-until-records-in-org"

    def __init__(self, sf, param):
        parts = self.split_pair(param)
        self.set_target_and_gap(*parts)
        query = f"select count(Id) from {self.sobject_name}"
        # TODO: fall back to other APIs
        self.in_org_count = sf.query(query)["records"][0]["expr0"]
        gap = self.target - int(self.in_org_count)
        self.gap = max(gap, 0)

    @property
    def nothing_to_do_because(self):
        return f"The org has {self.in_org_count} {self.sobject_name} records. You asked for {self.target}"


COUNT_STRATEGIES = {
    "run_until_recipe_repeated": RunUntilRecipeRepeated,
    "run_until_records_loaded": RunUntilRecordsLoaded,
    "run_until_records_in_org": RunUntilRecordInOrg,
}


def determine_run_until(options, sf):
    selected_strategies = [
        (strategy, options.get(strategy))
        for strategy in COUNT_STRATEGIES.keys()
        if options.get(strategy)
    ]

    if len(selected_strategies) > 1:
        raise exc.TaskOptionsError(
            "Please select only one of " + ", ".join(COUNT_STRATEGIES.keys())
        )
    elif not selected_strategies:
        strategy_choice = ("run_until_recipe_repeated", "1")
    else:
        strategy_choice = selected_strategies[0]

    strategy_name, param = strategy_choice
    strategy_impl = COUNT_STRATEGIES[strategy_name]

    return strategy_impl(sf, param)


class PortionGenerator:
    """
    Logic to generate batch sizes and know when we are done.

    Calling class must manage sets_created_so_far because it may
    have information unknown to this class, e.g. failures,
    processes running on other nodes, etc.
    """

    def __init__(
        self,
        target: int,
        min_batch_size: int,
        max_batch_size: int,
        start_count: int = 0,
    ):
        self.target = target
        self.min_batch_size = min_batch_size
        self.max_batch_size = max_batch_size
        self.next_batch_size = min_batch_size

    def done(self, sets_created_so_far):
        return self.gap(sets_created_so_far) <= 0

    def gap(self, sets_created_so_far):
        return self.target - sets_created_so_far

    def next_batch(self, sets_created_so_far: int):
        """Batch size starts at min_batch_size and grows toward max_batch_size
        unless the count gets there first

        The reason we grow is to try to engage the org's loader queue earlier
        than we would if we waited for the first big batch to be generated.
        """
        self.sets_created = sets_created_so_far
        # don't generate more records than needed to reach our target
        # or current batch size
        # or the max batch size
        self.batch_size = min(
            self.gap(sets_created_so_far), self.next_batch_size, self.max_batch_size
        )
        # don't generate fewer than zero
        self.batch_size = max(self.batch_size, 0)
        self.next_batch_size = int(self.next_batch_size * 1.1)
        return self.batch_size



import os
import sys

from simple_salesforce import api, bulk

from cumulusci.__about__ import __version__

__import__("pkg_resources").declare_namespace("cumulusci")

__location__ = os.path.dirname(os.path.realpath(__file__))

__version__ = __version__

if sys.version_info < (3, 8):  # pragma: no cover
    raise Exception("CumulusCI requires Python 3.8+.")

api.OrderedDict = dict
bulk.OrderedDict = dict
